from manim import *
from manim_ml.neural_network.layers.embedding import EmbeddingLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer


class FeedForwardToEmbedding(ConnectiveLayer):
    """Feed Forward to Embedding Layer"""

    input_class = FeedForwardLayer
    output_class = EmbeddingLayer

    def __init__(
        self,
        input_layer,
        output_layer,
        animation_dot_color=RED,
        dot_radius=0.03,
        **kwargs
    ):
        super().__init__(input_layer, output_layer, **kwargs)
        self.feed_forward_layer = input_layer
        self.embedding_layer = output_layer
        self.animation_dot_color = animation_dot_color
        self.dot_radius = dot_radius

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, layer_args={}, run_time=1.5, **kwargs):
        """Makes dots converge on a specific location"""
        # Find point to converge on by sampling from gaussian distribution
        location = self.embedding_layer.sample_point_location_from_distribution()
        # Set the embedding layer latent distribution
        # Move to location
        animations = []
        # Move the dots to the centers of each of the nodes in the FeedForwardLayer
        dots = []
        for node in self.feed_forward_layer.node_group:
            new_dot = Dot(
                node.get_center(),
                radius=self.dot_radius,
                color=self.animation_dot_color,
            )
            per_node_succession = Succession(
                Create(new_dot),
                new_dot.animate.move_to(location),
            )
            animations.append(per_node_succession)
            dots.append(new_dot)
        self.dots = VGroup(*dots)
        self.add(self.dots)
        # Follow up with remove animations
        remove_animations = []
        for dot in dots:
            remove_animations.append(FadeOut(dot))
        self.remove(self.dots)
        remove_animations = AnimationGroup(*remove_animations, run_time=0.2)
        animations = AnimationGroup(*animations)
        animation_group = Succession(animations, remove_animations, lag_ratio=1.0)

        return animation_group

    @override_animation(Create)
    def _create_override(self, **kwargs):
        return AnimationGroup()



from manim import *
from manim_ml.neural_network import NeuralNetwork, FeedForwardLayer, MathOperationLayer

# Make the specific scene
config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 6.0
config.frame_width = 6.0

class TestFeedForwardResidualNetwork(Scene):

    def construct(self):
        # Add the network
        nn = NeuralNetwork({
                "feed_forward_1": FeedForwardLayer(3),
                "feed_forward_2": FeedForwardLayer(3, activation_function="ReLU"),
                "feed_forward_3": FeedForwardLayer(3),
                "sum_operation": MathOperationLayer("+", activation_function="ReLU"),
            },
            layer_spacing=0.38
        )
        self.add(nn)
        # Make connections
        input_blank_dot = Dot(
            nn.input_layers_dict["feed_forward_1"].get_left() - np.array([0.65, 0.0, 0.0])
        )
        nn.add_connection(input_blank_dot, "feed_forward_1", arc_direction="straight")
        nn.add_connection("feed_forward_1", "sum_operation")
        output_blank_dot = Dot(
            nn.input_layers_dict["sum_operation"].get_right() + np.array([0.65, 0.0, 0.0])
        )
        nn.add_connection("sum_operation", output_blank_dot, arc_direction="straight")

        # Make forward pass animation
        self.play(
            nn.make_forward_pass_animation()
        )
    


from manim import *
from PIL import Image

from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.image import ImageLayer
from manim_ml.neural_network.neural_network import NeuralNetwork

# Make the specific scene
config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 6.0
config.frame_width = 6.0


class CombinedScene(ThreeDScene):
    def construct(self):
        image = Image.open("../assets/mnist/digit.jpeg")
        numpy_image = np.asarray(image)
        # Make nn
        nn = NeuralNetwork(
            [
                ImageLayer(numpy_image, height=1.5),
                Convolutional2DLayer(1, 7, 7, 3, 3, filter_spacing=0.32),
                Convolutional2DLayer(3, 5, 5, 3, 3, filter_spacing=0.32),
                FeedForwardLayer(3),
            ],
            layer_spacing=0.25,
        )
        # Center the nn
        nn.move_to(ORIGIN)
        nn.scale(1.3)
        self.add(nn)
        # Play animation
        forward_pass = nn.make_forward_pass_animation()
        self.wait(1)
        self.play(forward_pass)



"""
    Code for making a dropout animation for the
    feed forward layers of a neural network. 
"""
from manim import *
import random

from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.feed_forward_to_feed_forward import (
    FeedForwardToFeedForward,
)


class XMark(VGroup):
    def __init__(self, stroke_width=1.0, color=GRAY):
        super().__init__()
        line_one = Line(
            [-0.1, 0.1, 0],
            [0.1, -0.1, 0],
            stroke_width=1.0,
            stroke_color=color,
            z_index=4,
        )
        self.add(line_one)
        line_two = Line(
            [-0.1, -0.1, 0],
            [0.1, 0.1, 0],
            stroke_width=1.0,
            stroke_color=color,
            z_index=4,
        )
        self.add(line_two)


def get_edges_to_drop_out(layer: FeedForwardToFeedForward, layers_to_nodes_to_drop_out):
    """Returns edges to drop out for a given FeedForwardToFeedForward layer"""
    prev_layer = layer.input_layer
    next_layer = layer.output_layer
    # Get the nodes to dropout in previous layer
    prev_layer_nodes_to_dropout = layers_to_nodes_to_drop_out[prev_layer]
    next_layer_nodes_to_dropout = layers_to_nodes_to_drop_out[next_layer]
    # Compute the edges to dropout
    edges_to_dropout = []
    edge_indices_to_dropout = []
    for edge_index, edge in enumerate(layer.edges):
        prev_node_index = int(edge_index / next_layer.num_nodes)
        next_node_index = edge_index % next_layer.num_nodes
        # Check if the edges should be dropped out
        if (
            prev_node_index in prev_layer_nodes_to_dropout
            or next_node_index in next_layer_nodes_to_dropout
        ):
            edges_to_dropout.append(edge)
            edge_indices_to_dropout.append(edge_index)

    return edges_to_dropout, edge_indices_to_dropout


def make_pre_dropout_animation(
    neural_network,
    layers_to_nodes_to_drop_out,
    dropped_out_color=GRAY,
    dropped_out_opacity=0.2,
):
    """Makes an animation that sets up the NN layer for dropout"""
    animations = []
    # Go through the network and get the FeedForwardLayer instances
    feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardLayer)
    )
    # Go through the network and get the FeedForwardToFeedForwardLayer instances
    feed_forward_to_feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardToFeedForward)
    )
    # Get the edges to drop out
    layers_to_edges_to_dropout = {}
    for layer in feed_forward_to_feed_forward_layers:
        layers_to_edges_to_dropout[layer], _ = get_edges_to_drop_out(
            layer, layers_to_nodes_to_drop_out
        )
    # Dim the colors of the edges
    dim_edge_colors_animations = []
    for layer in layers_to_edges_to_dropout.keys():
        edges_to_drop_out = layers_to_edges_to_dropout[layer]
        # Make color dimming animation
        for edge_index, edge in enumerate(edges_to_drop_out):
            """
            def modify_edge(edge):
                edge.set_stroke_color(dropped_out_color)
                edge.set_stroke_width(0.6)
                edge.set_stroke_opacity(dropped_out_opacity)
                return edge

            dim_edge = ApplyFunction(
                modify_edge,
                edge
            )
            """

            dim_edge_colors_animations.append(FadeOut(edge))
    dim_edge_colors_animation = AnimationGroup(
        *dim_edge_colors_animations, lag_ratio=0.0
    )
    # Dim the colors of the nodes
    dim_nodes_animations = []
    x_marks = []
    for layer in layers_to_nodes_to_drop_out.keys():
        nodes_to_dropout = layers_to_nodes_to_drop_out[layer]
        # Make an X over each node
        for node_index, node in enumerate(layer.node_group):
            if node_index in nodes_to_dropout:
                x_mark = XMark()
                x_marks.append(x_mark)
                x_mark.move_to(node.get_center())
                create_x = Create(x_mark)
                dim_nodes_animations.append(create_x)

    dim_nodes_animation = AnimationGroup(*dim_nodes_animations, lag_ratio=0.0)

    animation_group = AnimationGroup(
        dim_edge_colors_animation,
        dim_nodes_animation,
    )

    return animation_group, x_marks


def make_post_dropout_animation(
    neural_network,
    layers_to_nodes_to_drop_out,
    x_marks,
):
    """Returns the NN to normal after dropout"""
    # Go through the network and get the FeedForwardLayer instances
    feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardLayer)
    )
    # Go through the network and get the FeedForwardToFeedForwardLayer instances
    feed_forward_to_feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardToFeedForward)
    )
    # Get the edges to drop out
    layers_to_edges_to_dropout = {}
    for layer in feed_forward_to_feed_forward_layers:
        layers_to_edges_to_dropout[layer], _ = get_edges_to_drop_out(
            layer, layers_to_nodes_to_drop_out
        )
    # Remove the x marks
    uncreate_animations = []
    for x_mark in x_marks:
        uncreate_x_mark = Uncreate(x_mark)
        uncreate_animations.append(uncreate_x_mark)

    uncreate_x_marks = AnimationGroup(*uncreate_animations, lag_ratio=0.0)
    # Add the edges back
    create_edge_animations = []
    for layer in layers_to_edges_to_dropout.keys():
        edges_to_drop_out = layers_to_edges_to_dropout[layer]
        # Make color dimming animation
        for edge_index, edge in enumerate(edges_to_drop_out):
            edge_copy = edge.copy()
            edges_to_drop_out[edge_index] = edge_copy
            create_edge_animations.append(FadeIn(edge_copy))

    create_edge_animation = AnimationGroup(*create_edge_animations, lag_ratio=0.0)

    return AnimationGroup(uncreate_x_marks, create_edge_animation, lag_ratio=0.0)


def make_forward_pass_with_dropout_animation(
    neural_network,
    layers_to_nodes_to_drop_out,
):
    """Makes forward pass animation with dropout"""
    layer_args = {}
    # Go through the network and get the FeedForwardLayer instances
    feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardLayer)
    )
    # Go through the network and get the FeedForwardToFeedForwardLayer instances
    feed_forward_to_feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardToFeedForward)
    )
    # Iterate through network and get feed forward layers
    for layer in feed_forward_layers:
        layer_args[layer] = {"dropout_node_indices": layers_to_nodes_to_drop_out[layer]}
    for layer in feed_forward_to_feed_forward_layers:
        _, edge_indices = get_edges_to_drop_out(layer, layers_to_nodes_to_drop_out)
        layer_args[layer] = {"edge_indices_to_dropout": edge_indices}

    return neural_network.make_forward_pass_animation(layer_args=layer_args)


def make_neural_network_dropout_animation(
    neural_network, dropout_rate=0.5, do_forward_pass=True, last_layer_stable=False, first_layer_stable=False, seed=None
):
    """
    Makes a dropout animation for a given neural network.

    NOTE Does dropout only on feed forward layers.

    1. Does dropout
    2. If `do_forward_pass` then do forward pass animation
    3. Revert network to pre-dropout appearance
    """
    # Go through the network and get the FeedForwardLayer instances
    if seed is not None:
        random.seed(seed)
    feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardLayer)
    )
    # Go through the network and get the FeedForwardToFeedForwardLayer instances
    feed_forward_to_feed_forward_layers = neural_network.filter_layers(
        lambda layer: isinstance(layer, FeedForwardToFeedForward)
    )
    # Get random nodes to drop out for each FeedForward Layer
    layers_to_nodes_to_drop_out = {}
    for idx, feed_forward_layer in enumerate(feed_forward_layers):
        num_nodes = feed_forward_layer.num_nodes
        nodes_to_drop_out = []
        # Compute random probability that each node is dropped out
        for node_index in range(num_nodes):
            dropout_prob = random.random()
            if last_layer_stable and idx==len(feed_forward_layers)-1:
                    continue
            if first_layer_stable and idx==0:
                continue
            if dropout_prob < dropout_rate:
                nodes_to_drop_out.append(node_index)
        # Add the mapping to the dict
        layers_to_nodes_to_drop_out[feed_forward_layer] = nodes_to_drop_out
    # Make the animation
    pre_dropout_animation, x_marks = make_pre_dropout_animation(
        neural_network, layers_to_nodes_to_drop_out
    )
    if do_forward_pass:
        forward_pass_animation = make_forward_pass_with_dropout_animation(
            neural_network, layers_to_nodes_to_drop_out
        )
    else:
        forward_pass_animation = AnimationGroup()

    post_dropout_animation = make_post_dropout_animation(
        neural_network, layers_to_nodes_to_drop_out, x_marks
    )
    # Combine the animations into one
    return Succession(
        pre_dropout_animation, forward_pass_animation, post_dropout_animation
    )



from manim import *
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer
from manim_ml.neural_network.layers.math_operation_layer import MathOperationLayer
from manim_ml.utils.mobjects.connections import NetworkConnection

class FeedForwardToMathOperation(ConnectiveLayer):
    """Image Layer to FeedForward layer"""

    input_class = FeedForwardLayer
    output_class = MathOperationLayer

    def __init__(
        self,
        input_layer,
        output_layer,
        active_color=ORANGE,
        **kwargs
    ):
        self.active_color = active_color
        super().__init__(input_layer, output_layer, **kwargs)

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        # Draw an arrow from the output of the feed forward layer to the
        # input of the math operation layer
        self.connection = NetworkConnection(
            self.input_layer,
            self.output_layer,
            arc_direction="straight",
            buffer=0.05
        )
        self.add(self.connection)

        return super().construct_layer(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Makes dots diverge from the given location and move to the feed forward nodes decoder"""
        # Make flashing pass animation on arrow
        passing_flash = ShowPassingFlash(
            self.connection.copy().set_color(self.active_color)
        )

        return passing_flash



from manim import *
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.utils.testing.frames_comparison import frames_comparison

from manim_ml.neural_network import NeuralNetwork, FeedForwardLayer, ImageLayer

from PIL import Image
import numpy as np

__module_test__ = "residual"


@frames_comparison
def test_ResidualConnectionScene(scene):
    """Tests the appearance of a residual connection"""
    nn = NeuralNetwork(
        {
            "layer1": FeedForwardLayer(3),
            "layer2": FeedForwardLayer(5),
            "layer3": FeedForwardLayer(3),
        }
    )

    scene.add(nn)


# Make the specific scene
config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 6.0
config.frame_width = 6.0


class FeedForwardScene(Scene):
    def construct(self):
        nn = NeuralNetwork(
            {
                "layer1": FeedForwardLayer(4),
                "layer2": FeedForwardLayer(4),
                "layer3": FeedForwardLayer(4),
            },
            layer_spacing=0.45,
        )

        nn.add_connection("layer1", "layer3")

        self.add(nn)

        self.play(nn.make_forward_pass_animation(), run_time=8)


class ConvScene(ThreeDScene):
    def construct(self):
        image = Image.open("../assets/mnist/digit.jpeg")
        numpy_image = np.asarray(image)

        nn = NeuralNetwork(
            {
                "layer1": Convolutional2DLayer(1, 5, padding=1),
                "layer2": Convolutional2DLayer(1, 5, 3, padding=1),
                "layer3": Convolutional2DLayer(1, 5, 3, padding=1),
            },
            layer_spacing=0.25,
        )

        nn.add_connection("layer1", "layer3")

        self.add(nn)

        self.play(nn.make_forward_pass_animation(), run_time=8)



from manim_ml.neural_network.activation_functions.relu import ReLUFunction
from manim_ml.neural_network.activation_functions.sigmoid import SigmoidFunction

name_to_activation_function_map = {"ReLU": ReLUFunction, "Sigmoid": SigmoidFunction}


def get_activation_function_by_name(name):
    assert (
        name in name_to_activation_function_map.keys()
    ), f"Unrecognized activation function {name}"

    return name_to_activation_function_map[name]



from manim import *

from manim_ml.neural_network.activation_functions.activation_function import (
    ActivationFunction,
)


class ReLUFunction(ActivationFunction):
    """Rectified Linear Unit Activation Function"""

    def __init__(self, function_name="ReLU", x_range=[-1, 1], y_range=[-1, 1]):
        super().__init__(function_name, x_range, y_range)

    def apply_function(self, x_val):
        if x_val < 0:
            return 0
        else:
            return x_val



import numpy as np

from manim import *
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer, ThreeDLayer
from manim_ml.utils.mobjects.gridded_rectangle import GriddedRectangle

from manim.utils.space_ops import rotation_matrix


def get_rotated_shift_vectors(input_layer, normalized=False):
    """Rotates the shift vectors"""
    # Make base shift vectors
    right_shift = np.array([input_layer.cell_width, 0, 0])
    down_shift = np.array([0, -input_layer.cell_width, 0])
    # Make rotation matrix
    rot_mat = rotation_matrix(ThreeDLayer.rotation_angle, ThreeDLayer.rotation_axis)
    # Rotate the vectors
    right_shift = np.dot(right_shift, rot_mat.T)
    down_shift = np.dot(down_shift, rot_mat.T)
    # Normalize the vectors
    if normalized:
        right_shift = right_shift / np.linalg.norm(right_shift)
        down_shift = down_shift / np.linalg.norm(down_shift)

    return right_shift, down_shift


class Filters(VGroup):
    """Group for showing a collection of filters connecting two layers"""

    def __init__(
        self,
        input_layer,
        output_layer,
        line_color=ORANGE,
        cell_width=1.0,
        stroke_width=2.0,
        show_grid_lines=False,
        output_feature_map_to_connect=None,  # None means all at once
    ):
        super().__init__()
        self.input_layer = input_layer
        self.output_layer = output_layer
        self.line_color = line_color
        self.cell_width = cell_width
        self.stroke_width = stroke_width
        self.show_grid_lines = show_grid_lines
        self.output_feature_map_to_connect = output_feature_map_to_connect
        # Make the filter
        self.input_rectangles = self.make_input_feature_map_rectangles()
        # self.input_rectangles.set_z_index(5)
        # self.add(self.input_rectangles)
        self.output_rectangles = self.make_output_feature_map_rectangles()
        # self.output_rectangles.set_z_index(5)
        # self.add(self.output_rectangles)
        self.connective_lines = self.make_connective_lines()
        # self.connective_lines.set_z_index(5)
        # self.add(self.connective_lines)

    def make_input_feature_map_rectangles(self):
        rectangles = []
        rectangle_width = (
            self.output_layer.filter_size[0] * self.output_layer.cell_width
        )
        rectangle_height = (
            self.output_layer.filter_size[1] * self.output_layer.cell_width
        )
        filter_color = self.output_layer.filter_color

        for index, feature_map in enumerate(self.input_layer.feature_maps):
            rectangle = GriddedRectangle(
                width=rectangle_width,
                height=rectangle_height,
                fill_color=filter_color,
                stroke_color=filter_color,
                fill_opacity=0.2,
                stroke_width=self.stroke_width,
                grid_xstep=self.cell_width,
                grid_ystep=self.cell_width,
                grid_stroke_width=self.stroke_width / 2,
                grid_stroke_color=filter_color,
                show_grid_lines=self.show_grid_lines,
            )
            # normal_vector = rectangle.get_normal_vector()
            rectangle.rotate(
                ThreeDLayer.rotation_angle,
                about_point=rectangle.get_center(),
                axis=ThreeDLayer.rotation_axis,
            )
            # Move the rectangle to the corner of the feature map
            rectangle.next_to(
                feature_map.get_corners_dict()["top_left"],
                submobject_to_align=rectangle.get_corners_dict()["top_left"],
                buff=0.0
                # aligned_edge=feature_map.get_corners_dict()["top_left"].get_center()
            )
            rectangle.set_z_index(5)

            rectangles.append(rectangle)

        feature_map_rectangles = VGroup(*rectangles)

        return feature_map_rectangles

    def make_output_feature_map_rectangles(self):
        rectangles = []

        rectangle_width = self.output_layer.cell_width
        rectangle_height = self.output_layer.cell_width
        filter_color = self.output_layer.filter_color
        right_shift, down_shift = get_rotated_shift_vectors(self.input_layer)
        left_shift = -1 * right_shift

        for index, feature_map in enumerate(self.output_layer.feature_maps):
            # Make sure current feature map is the right filter
            if not self.output_feature_map_to_connect is None:
                if index != self.output_feature_map_to_connect:
                    continue
            # Make the rectangle
            rectangle = GriddedRectangle(
                width=rectangle_width,
                height=rectangle_height,
                fill_color=filter_color,
                fill_opacity=0.2,
                stroke_color=filter_color,
                stroke_width=self.stroke_width,
                grid_xstep=self.cell_width,
                grid_ystep=self.cell_width,
                grid_stroke_width=self.stroke_width / 2,
                grid_stroke_color=filter_color,
                show_grid_lines=self.show_grid_lines,
            )
            # Rotate the rectangle
            rectangle.rotate(
                ThreeDLayer.rotation_angle,
                about_point=rectangle.get_center(),
                axis=ThreeDLayer.rotation_axis,
            )
            # Move the rectangle to the corner location
            rectangle.next_to(
                feature_map.get_corners_dict()["top_left"],
                submobject_to_align=rectangle.get_corners_dict()["top_left"],
                buff=0.0
                # aligned_edge=feature_map.get_corners_dict()["top_left"].get_center()
            )
            # Shift based on the amount of output layer padding
            rectangle.shift(
                self.output_layer.padding[0] * right_shift,
            )
            rectangle.shift(
                self.output_layer.padding[1] * down_shift,
            )
            rectangles.append(rectangle)

        feature_map_rectangles = VGroup(*rectangles)

        return feature_map_rectangles

    def make_connective_lines(self):
        """Lines connecting input filter with output node"""

        corner_names = ["top_left", "bottom_left", "top_right", "bottom_right"]

        def make_input_connective_lines():
            """Makes connective lines between the corners of the input filters"""
            first_input_rectangle = self.input_rectangles[0]
            last_input_rectangle = self.input_rectangles[-1]
            # Get the corner dots for each rectangle
            first_input_corners = first_input_rectangle.get_corners_dict()
            last_input_corners = last_input_rectangle.get_corners_dict()
            # Iterate through each corner and make the lines
            lines = []
            for corner_name in corner_names:
                line = Line(
                    first_input_corners[corner_name].get_center(),
                    last_input_corners[corner_name].get_center(),
                    color=self.line_color,
                    stroke_width=self.stroke_width,
                )
                lines.append(line)

            return VGroup(*lines)

        def make_output_connective_lines():
            """Makes connective lines between the corners of the output filters"""
            first_output_rectangle = self.output_rectangles[0]
            last_output_rectangle = self.output_rectangles[-1]
            # Get the corner dots for each rectangle
            first_output_corners = first_output_rectangle.get_corners_dict()
            last_output_corners = last_output_rectangle.get_corners_dict()
            # Iterate through each corner and make the lines
            lines = []
            for corner_name in corner_names:
                line = Line(
                    first_output_corners[corner_name].get_center(),
                    last_output_corners[corner_name].get_center(),
                    color=self.line_color,
                    stroke_width=self.stroke_width,
                )
                lines.append(line)

            return VGroup(*lines)

        def make_input_to_output_connective_lines():
            """Make connective lines between last input filter and first output filter"""
            # Choose the correct feature map to link to
            input_rectangle = self.input_rectangles[-1]
            output_rectangle = self.output_rectangles[0]
            # Get the corner dots for each rectangle
            input_corners = input_rectangle.get_corners_dict()
            output_corners = output_rectangle.get_corners_dict()
            # Iterate through each corner and make the lines
            lines = []
            for corner_name in corner_names:
                line = Line(
                    input_corners[corner_name].get_center(),
                    output_corners[corner_name].get_center(),
                    color=self.line_color,
                    stroke_width=self.stroke_width,
                )
                lines.append(line)

            return VGroup(*lines)

        input_lines = make_input_connective_lines()
        output_lines = make_output_connective_lines()
        input_output_lines = make_input_to_output_connective_lines()

        connective_lines = VGroup(*input_lines, *output_lines, *input_output_lines)

        return connective_lines

    @override_animation(Create)
    def _create_override(self, **kwargs):
        """
        NOTE This create override animation
        is a workaround to make sure that the filter
        does not show up in the scene before the create animation.

        Without this override the filters were shown at the beginning
        of the neural network forward pass animation
        instead of just when the filters were supposed to appear.
        I think this is a bug with Succession in the core
        Manim Community Library.

        TODO Fix this
        """

        def add_content(object):
            object.add(self.input_rectangles)
            object.add(self.connective_lines)
            object.add(self.output_rectangles)

            return object

        return ApplyFunction(add_content, self)
        return AnimationGroup(
            Create(self.input_rectangles),
            Create(self.connective_lines),
            Create(self.output_rectangles),
            lag_ratio=0.0,
        )

    def make_pulse_animation(self, shift_amount):
        """Make animation of the filter pulsing"""
        passing_flash = ShowPassingFlash(
            self.connective_lines.shift(shift_amount).set_stroke_width(
                self.stroke_width * 1.5
            ),
            time_width=0.2,
            color=RED,
            z_index=10,
        )

        return passing_flash


class Convolutional2DToConvolutional2D(ConnectiveLayer, ThreeDLayer):
    """Feed Forward to Embedding Layer"""

    input_class = Convolutional2DLayer
    output_class = Convolutional2DLayer

    def __init__(
        self,
        input_layer: Convolutional2DLayer,
        output_layer: Convolutional2DLayer,
        color=ORANGE,
        filter_opacity=0.3,
        line_color=ORANGE,
        active_color=ORANGE,
        cell_width=0.2,
        show_grid_lines=True,
        highlight_color=ORANGE,
        **kwargs,
    ):
        super().__init__(
            input_layer,
            output_layer,
            **kwargs,
        )
        self.color = color
        self.filter_color = self.output_layer.filter_color
        self.filter_size = self.output_layer.filter_size
        self.feature_map_size = self.input_layer.feature_map_size
        self.num_input_feature_maps = self.input_layer.num_feature_maps
        self.num_output_feature_maps = self.output_layer.num_feature_maps
        self.cell_width = self.output_layer.cell_width
        self.stride = self.output_layer.stride
        self.padding = self.input_layer.padding
        self.filter_opacity = filter_opacity
        self.cell_width = cell_width
        self.line_color = line_color
        self.active_color = active_color
        self.show_grid_lines = show_grid_lines
        self.highlight_color = highlight_color

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs,
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    def animate_filters_all_at_once(self, filters):
        """Animates each of the filters all at once"""
        animations = []
        # Make filters
        filters = Filters(
            self.input_layer,
            self.output_layer,
            line_color=self.color,
            cell_width=self.cell_width,
            show_grid_lines=self.show_grid_lines,
            output_feature_map_to_connect=None,  # None means all at once
        )
        animations.append(Create(filters))
        # Get the rotated shift vectors
        right_shift, down_shift = get_rotated_shift_vectors(self.input_layer)
        left_shift = -1 * right_shift
        # Make the animation
        num_y_moves = int(
            (self.feature_map_size[1] - self.filter_size[1]) / self.stride
            + self.padding[1] * 2
        )
        num_x_moves = int(
            (self.feature_map_size[0] - self.filter_size[0]) / self.stride
            + self.padding[0] * 2
        )
        for y_move in range(num_y_moves):
            # Go right num_x_moves
            for x_move in range(num_x_moves):
                # Shift right
                shift_animation = ApplyMethod(filters.shift, self.stride * right_shift)
                # shift_animation = self.animate.shift(right_shift)
                animations.append(shift_animation)
            # Go back left num_x_moves and down one
            shift_amount = (
                self.stride * num_x_moves * left_shift + self.stride * down_shift
            )
            # Make the animation
            shift_animation = ApplyMethod(filters.shift, shift_amount)
            animations.append(shift_animation)
        # Do last row move right
        for x_move in range(num_x_moves):
            # Shift right
            shift_animation = ApplyMethod(filters.shift, self.stride * right_shift)
            # shift_animation = self.animate.shift(right_shift)
            animations.append(shift_animation)
        # Remove the filters
        animations.append(FadeOut(filters))
        return Succession(*animations, lag_ratio=1.0)

    def animate_filters_one_at_a_time(self, highlight_active_feature_map=True):
        """Animates each of the filters one at a time"""
        animations = []
        output_feature_maps = self.output_layer.feature_maps
        for feature_map_index in range(len(output_feature_maps)):
            # Make filters
            filters = Filters(
                self.input_layer,
                self.output_layer,
                line_color=self.color,
                cell_width=self.cell_width,
                show_grid_lines=self.show_grid_lines,
                output_feature_map_to_connect=feature_map_index,  # None means all at once
            )
            animations.append(Create(filters))
            # Highlight the feature map
            if highlight_active_feature_map:
                feature_map = output_feature_maps[feature_map_index]
                original_feature_map_color = feature_map.color
                # Change the output feature map colors
                change_color_animations = []
                change_color_animations.append(
                    ApplyMethod(feature_map.set_color, self.highlight_color)
                )
                # Change the input feature map colors
                input_feature_maps = self.input_layer.feature_maps
                for input_feature_map in input_feature_maps:
                    change_color_animations.append(
                        ApplyMethod(input_feature_map.set_color, self.highlight_color)
                    )
                # Combine the animations
                animations.append(
                    AnimationGroup(*change_color_animations, lag_ratio=0.0)
                )
            # Get the rotated shift vectors
            right_shift, down_shift = get_rotated_shift_vectors(self.input_layer)
            left_shift = -1 * right_shift
            # Make the animation
            num_y_moves = int(
                (self.feature_map_size[1] - self.filter_size[1]) / self.stride
                + self.padding[1] * 2
            )
            num_x_moves = int(
                (self.feature_map_size[0] - self.filter_size[0]) / self.stride
                + self.padding[0] * 2
            )
            for y_move in range(num_y_moves):
                # Go right num_x_moves
                for x_move in range(num_x_moves):
                    # Make a pulse animation for the corners
                    """
                    pulse_animation = filters.make_pulse_animation(
                        shift_amount=shift_amount
                    )
                    animations.append(pulse_animation)
                    """
                    z_index_animation = ApplyMethod(filters.set_z_index, 5)
                    animations.append(z_index_animation)
                    # Shift right
                    shift_animation = ApplyMethod(
                        filters.shift, self.stride * right_shift
                    )
                    # shift_animation = self.animate.shift(right_shift)
                    animations.append(shift_animation)

                # Go back left num_x_moves and down one
                shift_amount = (
                    self.stride * num_x_moves * left_shift + self.stride * down_shift
                )
                # Make the animation
                shift_animation = ApplyMethod(filters.shift, shift_amount)
                animations.append(shift_animation)
            # Do last row move right
            for x_move in range(num_x_moves):
                # Shift right
                shift_animation = ApplyMethod(filters.shift, self.stride * right_shift)
                # shift_animation = self.animate.shift(right_shift)
                animations.append(shift_animation)
            # Remove the filters
            animations.append(FadeOut(filters))
            # Un-highlight the feature map
            if highlight_active_feature_map:
                feature_map = output_feature_maps[feature_map_index]
                # Change the output feature map colors
                change_color_animations = []
                change_color_animations.append(
                    ApplyMethod(feature_map.set_color, original_feature_map_color)
                )
                # Change the input feature map colors
                input_feature_maps = self.input_layer.feature_maps
                for input_feature_map in input_feature_maps:
                    change_color_animations.append(
                        ApplyMethod(
                            input_feature_map.set_color, original_feature_map_color
                        )
                    )
                # Combine the animations
                animations.append(
                    AnimationGroup(*change_color_animations, lag_ratio=0.0)
                )

        return Succession(*animations, lag_ratio=1.0)

    def make_forward_pass_animation(
        self,
        layer_args={},
        all_filters_at_once=False,
        highlight_active_feature_map=True,
        run_time=10.5,
        **kwargs,
    ):
        """Forward pass animation from conv2d to conv2d"""
        print(f"All filters at once: {all_filters_at_once}")
        # Make filter shifting animations
        if all_filters_at_once:
            return self.animate_filters_all_at_once()
        else:
            return self.animate_filters_one_at_a_time(
                highlight_active_feature_map=highlight_active_feature_map
            )

    def scale(self, scale_factor, **kwargs):
        self.cell_width *= scale_factor
        super().scale(scale_factor, **kwargs)

    @override_animation(Create)
    def _create_override(self, **kwargs):
        return Succession()



from manim import *

from manim_ml.neural_network.activation_functions import get_activation_function_by_name
from manim_ml.neural_network.activation_functions.activation_function import (
    ActivationFunction,
)
from manim_ml.neural_network.layers.parent_layers import VGroupNeuralNetworkLayer

class MathOperationLayer(VGroupNeuralNetworkLayer):
    """Handles rendering a layer for a neural network"""
    valid_operations = ["+", "-", "*", "/"]

    def __init__(
        self,
        operation_type: str,
        node_radius=0.5,
        node_color=BLUE,
        node_stroke_width=2.0,
        active_color=ORANGE,
        activation_function=None,
        font_size=20,
        **kwargs
    ):
        super(VGroupNeuralNetworkLayer, self).__init__(**kwargs)
        # Ensure operation type is valid
        assert operation_type in MathOperationLayer.valid_operations
        self.operation_type = operation_type
        self.node_radius = node_radius
        self.node_color = node_color
        self.node_stroke_width = node_stroke_width
        self.active_color = active_color
        self.font_size = font_size
        self.activation_function = activation_function

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        """Creates the neural network layer"""
        # Draw the operation
        self.operation_text = Text(
            self.operation_type,
            font_size=self.font_size
        )
        self.add(self.operation_text)
        # Make the surrounding circle
        self.surrounding_circle = Circle(
            color=self.node_color,
            stroke_width=self.node_stroke_width
        ).surround(self.operation_text)
        self.add(self.surrounding_circle)
        # Make the activation function
        self.construct_activation_function()
        super().construct_layer(input_layer, output_layer, **kwargs)

    def construct_activation_function(self):
        """Construct the activation function"""
        # Add the activation function
        if not self.activation_function is None:
            # Check if it is a string
            if isinstance(self.activation_function, str):
                activation_function = get_activation_function_by_name(
                    self.activation_function
                )()
            else:
                assert isinstance(self.activation_function, ActivationFunction)
                activation_function = self.activation_function
            # Plot the function above the rest of the layer
            self.activation_function = activation_function
            self.add(self.activation_function)

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Makes the forward pass animation

        Parameters
        ----------
        layer_args : dict, optional
            layer specific arguments, by default {}

        Returns
        -------
        AnimationGroup
            Forward pass animation
        """
        # Make highlight animation
        succession = Succession(
            ApplyMethod(
                self.surrounding_circle.set_color, 
                self.active_color, 
                run_time=0.25
            ),
            Wait(1.0),
            ApplyMethod(
                self.surrounding_circle.set_color, 
                self.node_color, 
                run_time=0.25
            ),
        )
        # Animate the activation function
        if not self.activation_function is None:
            animation_group = AnimationGroup(
                succession,
                self.activation_function.make_evaluate_animation(),
                lag_ratio=0.0,
            )
            return animation_group
        else:
            return succession

    def get_center(self):
        return self.surrounding_circle.get_center()

    def get_left(self):
        return self.surrounding_circle.get_left()

    def get_right(self):
        return self.surrounding_circle.get_right()
    
    def move_to(self, mobject_or_point):
        """Moves the center of the layer to the given mobject or point"""
        layer_center = self.surrounding_circle.get_center()
        if isinstance(mobject_or_point, Mobject):
            target_center = mobject_or_point.get_center() 
        else:
            target_center = mobject_or_point

        self.shift(target_center - layer_center)


from manim import *

from manim_ml.neural_network.activation_functions import get_activation_function_by_name
from manim_ml.neural_network.activation_functions.activation_function import (
    ActivationFunction,
)
from manim_ml.neural_network.layers.parent_layers import VGroupNeuralNetworkLayer
import manim_ml

class FeedForwardLayer(VGroupNeuralNetworkLayer):
    """Handles rendering a layer for a neural network"""

    def __init__(
        self,
        num_nodes,
        layer_buffer=SMALL_BUFF / 2,
        node_radius=0.08,
        node_color=manim_ml.config.color_scheme.primary_color,
        node_outline_color=manim_ml.config.color_scheme.secondary_color,
        rectangle_color=manim_ml.config.color_scheme.secondary_color,
        node_spacing=0.3,
        rectangle_fill_color=manim_ml.config.color_scheme.background_color,
        node_stroke_width=2.0,
        rectangle_stroke_width=2.0,
        animation_dot_color=manim_ml.config.color_scheme.active_color,
        activation_function=None,
        **kwargs
    ):
        super(VGroupNeuralNetworkLayer, self).__init__(**kwargs)
        self.num_nodes = num_nodes
        self.layer_buffer = layer_buffer
        self.node_radius = node_radius
        self.node_color = node_color
        self.node_stroke_width = node_stroke_width
        self.node_outline_color = node_outline_color
        self.rectangle_stroke_width = rectangle_stroke_width
        self.rectangle_color = rectangle_color
        self.node_spacing = node_spacing
        self.rectangle_fill_color = rectangle_fill_color
        self.animation_dot_color = animation_dot_color
        self.activation_function = activation_function

        self.node_group = VGroup()

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        """Creates the neural network layer"""
        # Add Nodes
        for node_number in range(self.num_nodes):
            node_object = Circle(
                radius=self.node_radius,
                color=self.node_color,
                stroke_width=self.node_stroke_width,
            )
            self.node_group.add(node_object)
        # Space the nodes
        # Assumes Vertical orientation
        for node_index, node_object in enumerate(self.node_group):
            location = node_index * self.node_spacing
            node_object.move_to([0, location, 0])
        # Create Surrounding Rectangle
        self.surrounding_rectangle = SurroundingRectangle(
            self.node_group,
            color=self.rectangle_color,
            fill_color=self.rectangle_fill_color,
            fill_opacity=1.0,
            buff=self.layer_buffer,
            stroke_width=self.rectangle_stroke_width,
        )
        self.surrounding_rectangle.set_z_index(1)
        # Add the objects to the class
        self.add(self.surrounding_rectangle, self.node_group)

        self.construct_activation_function()
        super().construct_layer(input_layer, output_layer, **kwargs)

    def construct_activation_function(self):
        """Construct the activation function"""
        # Add the activation function
        if not self.activation_function is None:
            # Check if it is a string
            if isinstance(self.activation_function, str):
                activation_function = get_activation_function_by_name(
                    self.activation_function
                )()
            else:
                assert isinstance(self.activation_function, ActivationFunction)
                activation_function = self.activation_function
            # Plot the function above the rest of the layer
            self.activation_function = activation_function
            self.add(self.activation_function)

    def make_dropout_forward_pass_animation(self, layer_args, **kwargs):
        """Makes a forward pass animation with dropout"""
        # Make sure proper dropout information was passed
        assert "dropout_node_indices" in layer_args
        dropout_node_indices = layer_args["dropout_node_indices"]
        # Only highlight nodes that were note dropped out
        nodes_to_highlight = []
        for index, node in enumerate(self.node_group):
            if not index in dropout_node_indices:
                nodes_to_highlight.append(node)
        nodes_to_highlight = VGroup(*nodes_to_highlight)
        # Make highlight animation
        succession = Succession(
            ApplyMethod(
                nodes_to_highlight.set_color, self.animation_dot_color, run_time=0.25
            ),
            Wait(1.0),
            ApplyMethod(nodes_to_highlight.set_color, self.node_color, run_time=0.25),
        )

        return succession

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        # Check if dropout is a thing
        if "dropout_node_indices" in layer_args:
            # Drop out certain nodes
            return self.make_dropout_forward_pass_animation(
                layer_args=layer_args, **kwargs
            )
        else:
            # Make highlight animation
            succession = Succession(
                ApplyMethod(
                    self.node_group.set_color, self.animation_dot_color, run_time=0.25
                ),
                Wait(1.0),
                ApplyMethod(self.node_group.set_color, self.node_color, run_time=0.25),
            )
            if not self.activation_function is None:
                animation_group = AnimationGroup(
                    succession,
                    self.activation_function.make_evaluate_animation(),
                    lag_ratio=0.0,
                )
                return animation_group
            else:
                return succession

    @override_animation(Create)
    def _create_override(self, **kwargs):
        animations = []

        animations.append(Create(self.surrounding_rectangle))

        for node in self.node_group:
            animations.append(Create(node))

        animation_group = AnimationGroup(*animations, lag_ratio=0.0)
        return animation_group

    def get_height(self):
        return self.surrounding_rectangle.get_height()

    def get_center(self):
        return self.surrounding_rectangle.get_center()

    def get_left(self):
        return self.surrounding_rectangle.get_left()

    def get_right(self):
        return self.surrounding_rectangle.get_right()

    def move_to(self, mobject_or_point):
        """Moves the center of the layer to the given mobject or point"""
        layer_center = self.surrounding_rectangle.get_center()
        if isinstance(mobject_or_point, Mobject):
            target_center = mobject_or_point.get_center() 
        else:
            target_center = mobject_or_point

        self.shift(target_center - layer_center)


from manim import *
import random

from manim_ml.neural_network.layers.parent_layers import VGroupNeuralNetworkLayer


class VectorLayer(VGroupNeuralNetworkLayer):
    """Shows a vector"""

    def __init__(self, num_values, value_func=lambda: random.uniform(0, 1), **kwargs):
        super().__init__(**kwargs)
        self.num_values = num_values
        self.value_func = value_func

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs,
    ):
        super().construct_layer(input_layer, output_layer, **kwargs)
        # Make the vector
        self.vector_label = self.make_vector()
        self.add(self.vector_label)

    def make_vector(self):
        """Makes the vector"""
        if False:
            # TODO install Latex
            values = np.array([self.value_func() for i in range(self.num_values)])
            values = values[None, :].T
            vector = Matrix(values)

        vector_label = Text(f"[{self.value_func():.2f}]")
        vector_label.scale(0.3)

        return vector_label

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        return AnimationGroup()

    @override_animation(Create)
    def _create_override(self):
        """Create animation"""
        return Write(self.vector_label)



from manim import *
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer, ThreeDLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer


class Convolutional2DToFeedForward(ConnectiveLayer, ThreeDLayer):
    """Feed Forward to Embedding Layer"""

    input_class = Convolutional2DLayer
    output_class = FeedForwardLayer

    def __init__(
        self,
        input_layer: Convolutional2DLayer,
        output_layer: FeedForwardLayer,
        passing_flash_color=ORANGE,
        **kwargs
    ):
        super().__init__(input_layer, output_layer, **kwargs)
        self.passing_flash_color = passing_flash_color

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, layer_args={}, run_time=1.5, **kwargs):
        """Forward pass animation from conv2d to conv2d"""
        animations = []
        # Get input layer final feature map
        final_feature_map = self.input_layer.feature_maps[-1]
        # Get output layer nodes
        feed_forward_nodes = self.output_layer.node_group
        # Go through each corner
        corners = final_feature_map.get_corners_dict().values()
        for corner in corners:
            # Go through each node
            for node in feed_forward_nodes:
                line = Line(corner, node, stroke_width=1.0)
                line.set_z_index(self.output_layer.node_group.get_z_index())
                anim = ShowPassingFlash(
                    line.set_color(self.passing_flash_color), time_width=0.2
                )
                animations.append(anim)

        return AnimationGroup(*animations)



from manim import *

from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.neural_network import NeuralNetwork

from manim_ml.utils.testing.frames_comparison import frames_comparison

__module_test__ = "padding"

# Make the specific scene
config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 6.0
config.frame_width = 6.0


class CombinedScene(ThreeDScene):
    def construct(self):
        # Make nn
        nn = NeuralNetwork(
            [
                Convolutional2DLayer(
                    num_feature_maps=1,
                    feature_map_size=7,
                    padding=1,
                    padding_dashed=True,
                ),
                Convolutional2DLayer(
                    num_feature_maps=3,
                    feature_map_size=7,
                    filter_size=3,
                    padding=0,
                    padding_dashed=False,
                ),
                FeedForwardLayer(3),
            ],
            layer_spacing=0.25,
        )
        # Center the nn
        nn.move_to(ORIGIN)
        self.add(nn)
        # Play animation
        forward_pass = nn.make_forward_pass_animation()
        self.wait(1)
        self.play(forward_pass, run_time=30)


@frames_comparison
def test_ConvPadding(scene):
    # Make nn
    nn = NeuralNetwork(
        [
            Convolutional2DLayer(
                num_feature_maps=1, feature_map_size=7, padding=1, padding_dashed=True
            ),
            Convolutional2DLayer(
                num_feature_maps=3,
                feature_map_size=7,
                filter_size=3,
                padding=1,
                filter_spacing=0.35,
                padding_dashed=False,
            ),
            FeedForwardLayer(3),
        ],
        layer_spacing=0.25,
    )
    # Center the nn
    nn.move_to(ORIGIN)
    scene.add(nn)
    # Play animation
    forward_pass = nn.make_forward_pass_animation()
    scene.play(forward_pass, run_time=30)



from manim import *
from manim_ml.utils.mobjects.probability import GaussianDistribution


class TestShowGaussian(Scene):
    def construct(self):
        axes = Axes()
        self.add(axes)
        gaussian = GaussianDistribution(
            axes, mean=np.array([0.0, 0.0]), cov=np.array([[2.0, 0.0], [0.0, 1.0]])
        )
        self.add(gaussian)



from manim import *
from manim_ml.decision_tree.decision_tree import DecisionTreeDiagram
from sklearn.tree import DecisionTreeClassifier
from sklearn import datasets
import sklearn
import matplotlib.pyplot as plt


def learn_iris_decision_tree(iris):
    decision_tree = DecisionTreeClassifier(
        random_state=1, max_depth=3, max_leaf_nodes=6
    )
    decision_tree = decision_tree.fit(iris.data[:, :2], iris.target)
    # output the decisioin tree in some format
    return decision_tree


def make_sklearn_tree(dataset, max_tree_depth=3):
    tree = learn_iris_decision_tree(dataset)
    feature_names = dataset.feature_names[0:2]
    return tree, tree.tree_


class DecisionTreeScene(Scene):

    def construct(self):
        """Makes a decision tree object"""
        iris_dataset = datasets.load_iris()
        clf, sklearn_tree = make_sklearn_tree(iris_dataset)
        # sklearn.tree.plot_tree(clf, node_ids=True)
        # plt.show()
        decision_tree = DecisionTreeDiagram(
            sklearn_tree,
            class_images_paths=[
                "images/iris_dataset/SetosaFlower.jpeg",
                "images/iris_dataset/VeriscolorFlower.jpeg",
                "images/iris_dataset/VirginicaFlower.jpeg",
            ],
            class_names=["Setosa", "Veriscolor", "Virginica"],
            feature_names=["Sepal Length", "Sepal Width"],
        )
        decision_tree.move_to(ORIGIN)
        create_decision_tree = Create(decision_tree, traversal_order="bfs")
        self.play(create_decision_tree)
        # self.play(create_decision_tree)

class SurfacePlot(Scene):

    def construct(self):
        iris_dataset = datasets.load_iris()
        iris_dataset_plot = IrisDatasetPlot(iris_dataset)
        iris_dataset_plot.all_group.scale(1.0)
        iris_dataset_plot.all_group.shift([-3, 0.2, 0])
        self.play(Create(iris_dataset_plot))
        # make the decision tree classifier
        decision_tree_classifier, sklearn_tree = make_sklearn_tree(iris_dataset)
        decision_tree_surface = DecisionTreeSurface(
            decision_tree_classifier, iris_dataset.data, iris_dataset_plot.axes_group[0]
        )
        self.play(Create(decision_tree_surface))
        self.wait(1)



import numpy as np

from manim import *
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.neural_network.layers.image import ImageLayer
from manim_ml.neural_network.layers.parent_layers import (
    ThreeDLayer,
    VGroupNeuralNetworkLayer,
)
from manim_ml.utils.mobjects.gridded_rectangle import GriddedRectangle


class ImageToConvolutional2DLayer(VGroupNeuralNetworkLayer, ThreeDLayer):
    """Handles rendering a convolutional layer for a nn"""

    input_class = ImageLayer
    output_class = Convolutional2DLayer

    def __init__(
        self, input_layer: ImageLayer, output_layer: Convolutional2DLayer, **kwargs
    ):
        super().__init__(input_layer, output_layer, **kwargs)
        self.input_layer = input_layer
        self.output_layer = output_layer

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs,
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, run_time=5, layer_args={}, **kwargs):
        """Maps image to convolutional layer"""
        # Transform the image from the input layer to the
        num_image_channels = self.input_layer.num_channels
        if num_image_channels == 1 or num_image_channels == 3:  # TODO fix this later
            return self.grayscale_image_forward_pass_animation()
        elif num_image_channels == 3:
            return self.rbg_image_forward_pass_animation()
        else:
            raise Exception(
                f"Unrecognized number of image channels: {num_image_channels}"
            )

    def rbg_image_forward_pass_animation(self):
        """Handles forward pass animation for 3 channel image"""
        image_mobject = self.input_layer.image_mobject
        # TODO get each color channel and turn it into an image
        # TODO create image mobjects for each channel and transform
        # it to the feature maps of the output_layer
        raise NotImplementedError()

    def grayscale_image_forward_pass_animation(self):
        """Handles forward pass animation for 1 channel image"""
        animations = []
        image_mobject = self.input_layer.image_mobject
        target_feature_map = self.output_layer.feature_maps[0]
        # Map image mobject to feature map
        # Make rotation of image
        rotation = ApplyMethod(
            image_mobject.rotate,
            ThreeDLayer.rotation_angle,
            ThreeDLayer.rotation_axis,
            image_mobject.get_center(),
            run_time=0.5,
        )
        """
        x_rotation = ApplyMethod(
            image_mobject.rotate,
            ThreeDLayer.three_d_x_rotation,
            [1, 0, 0], 
            image_mobject.get_center(),
            run_time=0.5
        )
        y_rotation = ApplyMethod(
            image_mobject.rotate,
            ThreeDLayer.three_d_y_rotation,
            [0, 1, 0], 
            image_mobject.get_center(),
            run_time=0.5
        )
        """
        # Set opacity
        set_opacity = ApplyMethod(image_mobject.set_opacity, 0.2, run_time=0.5)
        # Scale the max of width or height to the
        # width of the feature_map
        def scale_image_func(image_mobject):
            max_width_height = max(image_mobject.width, image_mobject.height)
            scale_factor = target_feature_map.untransformed_width / max_width_height
            image_mobject.scale(scale_factor)

            return image_mobject

        scale_image = ApplyFunction(scale_image_func, image_mobject)
        # scale_image = ApplyMethod(image_mobject.scale, scale_factor, run_time=0.5)
        # Move the image
        move_image = ApplyMethod(image_mobject.move_to, target_feature_map)
        # Compose the animations
        animation = Succession(
            rotation,
            scale_image,
            set_opacity,
            move_image,
        )
        return animation

    def scale(self, scale_factor, **kwargs):
        super().scale(scale_factor, **kwargs)

    @override_animation(Create)
    def _create_override(self, **kwargs):
        return AnimationGroup()




from manim import *

import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib
plt.style.use('dark_background')

from manim_ml.utils.mobjects.plotting import convert_matplotlib_figure_to_image_mobject
from manim_ml.utils.testing.frames_comparison import frames_comparison

__module_test__ = "plotting"

@frames_comparison
def test_matplotlib_to_image_mobject(scene):
    # libraries & dataset
    df = sns.load_dataset('iris')
    # Custom the color, add shade and bandwidth
    matplotlib.use('Agg')
    plt.figure(figsize=(10,10), dpi=100)
    displot = sns.displot(
        x=df.sepal_width, 
        y=df.sepal_length, 
        cmap="Reds", 
        kind="kde",
    )
    plt.axis('off')
    fig = displot.fig
    image_mobject = convert_matplotlib_figure_to_image_mobject(fig)
    # Display the image mobject
    scene.add(image_mobject)

class TestMatplotlibToImageMobject(Scene):

    def construct(self):
        # Make a matplotlib plot
        # libraries & dataset
        df = sns.load_dataset('iris')
        # Custom the color, add shade and bandwidth
        matplotlib.use('Agg')
        plt.figure(figsize=(10,10), dpi=100)
        displot = sns.displot(
            x=df.sepal_width, 
            y=df.sepal_length, 
            cmap="Reds", 
            kind="kde",
        )
        plt.axis('off')
        fig = displot.fig
        image_mobject = convert_matplotlib_figure_to_image_mobject(fig)
        # Display the image mobject
        self.add(image_mobject)


class HexabinScene(Scene):

    def construct(self):
        # Fixing random state for reproducibility
        np.random.seed(19680801)
        n = 100_000
        x = np.random.standard_normal(n)
        y = x + 1.0 * np.random.standard_normal(n)
        xlim = -4, 4
        ylim = -4, 4

        fig, ax0 = plt.subplots(1, figsize=(5, 5))

        hb = ax0.hexbin(x, y, gridsize=50, cmap='inferno')
        ax0.set(xlim=xlim, ylim=ylim)

        self.add(convert_matplotlib_figure_to_image_mobject(fig))



from manim import *
import numpy as np
from PIL import Image

class GrayscaleImageMobject(Group):
    """Mobject for creating images in Manim from numpy arrays"""

    def __init__(self, numpy_image, height=2.3):
        super().__init__()
        self.numpy_image = numpy_image
        assert len(np.shape(self.numpy_image)) == 2

        input_image = self.numpy_image[None, :, :]
        # Convert grayscale to rgb version of grayscale
        input_image = np.repeat(input_image, 3, axis=0)
        input_image = np.rollaxis(input_image, 0, start=3)
        self.image_mobject = ImageMobject(
            input_image, 
            image_mode="RBG",
        )
        self.add(self.image_mobject)
        self.image_mobject.set_resampling_algorithm(
            RESAMPLING_ALGORITHMS["nearest"]
        )
        self.image_mobject.scale_to_fit_height(height)

    @classmethod
    def from_path(cls, path, height=2.3):
        """Loads image from path"""
        image = Image.open(path)
        numpy_image = np.asarray(image)

        return cls(numpy_image, height=height)

    @override_animation(Create)
    def create(self, run_time=2):
        return FadeIn(self)

    def scale(self, scale_factor, **kwargs):
        """Scales the image mobject"""
        # super().scale(scale_factor)
        # height = self.height
        self.image_mobject.scale(scale_factor)
        # self.scale_to_fit_height(2)
        # self.apply_points_function_about_point(
        #     lambda points: scale_factor * points, **kwargs
        # )

    def set_opacity(self, opacity):
        """Set the opacity"""
        self.image_mobject.set_opacity(opacity)


class LabeledColorImage(Group):
    """Labeled Color Image"""

    def __init__(
        self, image, color=RED, label="Positive", stroke_width=5, font_size=24, buff=0.2
    ):
        super().__init__()
        self.image = image
        self.color = color
        self.label = label
        self.stroke_width = stroke_width
        self.font_size = font_size

        text = Text(label, font_size=self.font_size)
        text.next_to(self.image, UP, buff=buff)
        rectangle = SurroundingRectangle(
            self.image, color=color, buff=0.0, stroke_width=self.stroke_width
        )

        self.add(text)
        self.add(rectangle)
        self.add(self.image)



import manim
from manim_ml.utils.colorschemes.colorschemes import light_mode, dark_mode, ColorScheme

class ManimMLConfig:

    def __init__(self, default_color_scheme=light_mode):
        self._color_scheme = default_color_scheme

    @property
    def color_scheme(self):
        return self._color_scheme
    
    @color_scheme.setter
    def color_scheme(self, value):
        if isinstance(value, str):
            if value == "dark_mode":
                self._color_scheme = dark_mode
            elif value == "light_mode":
                self._color_scheme = light_mode
            else:
                raise ValueError(
                    "Color scheme must be either 'dark_mode' or 'light_mode'"
                )
        elif isinstance(value, ColorScheme):
            self._color_scheme = value
            
        manim.config.background_color = self.color_scheme.background_color

# These are accesible from the manim_ml namespace
config = ManimMLConfig()


from typing import Union
from manim_ml.neural_network.activation_functions import get_activation_function_by_name
from manim_ml.neural_network.activation_functions.activation_function import (
    ActivationFunction,
)
import numpy as np
from manim import *

from manim_ml.neural_network.layers.parent_layers import (
    ThreeDLayer,
    VGroupNeuralNetworkLayer,
)
from manim_ml.utils.mobjects.gridded_rectangle import GriddedRectangle


class FeatureMap(VGroup):
    """Class for making a feature map"""

    def __init__(
        self,
        color=ORANGE,
        feature_map_size=None,
        fill_color=ORANGE,
        fill_opacity=0.2,
        cell_width=0.2,
        padding=(0, 0),
        stroke_width=2.0,
        show_grid_lines=False,
        padding_dashed=False,
    ):
        super().__init__()
        self.color = color
        self.feature_map_size = feature_map_size
        self.fill_color = fill_color
        self.fill_opacity = fill_opacity
        self.cell_width = cell_width
        self.padding = padding
        self.stroke_width = stroke_width
        self.show_grid_lines = show_grid_lines
        self.padding_dashed = padding_dashed
        # Check if we have non-zero padding
        if padding[0] > 0 or padding[1] > 0:
            # Make the exterior rectangle dashed
            width_with_padding = (
                self.feature_map_size[0] + self.padding[0] * 2
            ) * self.cell_width
            height_with_padding = (
                self.feature_map_size[1] + self.padding[1] * 2
            ) * self.cell_width
            self.untransformed_width = width_with_padding
            self.untransformed_height = height_with_padding

            self.exterior_rectangle = GriddedRectangle(
                color=self.color,
                width=width_with_padding,
                height=height_with_padding,
                fill_color=self.color,
                fill_opacity=self.fill_opacity,
                stroke_color=self.color,
                stroke_width=self.stroke_width,
                grid_xstep=self.cell_width,
                grid_ystep=self.cell_width,
                grid_stroke_width=self.stroke_width / 2,
                grid_stroke_color=self.color,
                show_grid_lines=self.show_grid_lines,
                dotted_lines=self.padding_dashed,
            )
            self.add(self.exterior_rectangle)
            # Add an interior rectangle with no fill color
            self.interior_rectangle = GriddedRectangle(
                color=self.color,
                fill_opacity=0.0,
                width=self.feature_map_size[0] * self.cell_width,
                height=self.feature_map_size[1] * self.cell_width,
                stroke_width=self.stroke_width,
            )
            self.add(self.interior_rectangle)
        else:
            # Just make an exterior rectangle with no dashes.
            self.untransformed_height = (self.feature_map_size[1] * self.cell_width,)
            self.untransformed_width = (self.feature_map_size[0] * self.cell_width,)
            # Make the exterior rectangle
            self.exterior_rectangle = GriddedRectangle(
                color=self.color,
                height=self.feature_map_size[1] * self.cell_width,
                width=self.feature_map_size[0] * self.cell_width,
                fill_color=self.color,
                fill_opacity=self.fill_opacity,
                stroke_color=self.color,
                stroke_width=self.stroke_width,
                grid_xstep=self.cell_width,
                grid_ystep=self.cell_width,
                grid_stroke_width=self.stroke_width / 2,
                grid_stroke_color=self.color,
                show_grid_lines=self.show_grid_lines,
            )
            self.add(self.exterior_rectangle)

    def get_corners_dict(self):
        """Returns a dictionary of the corners"""
        # Sort points through clockwise rotation of a vector in the xy plane
        return self.exterior_rectangle.get_corners_dict()


class Convolutional2DLayer(VGroupNeuralNetworkLayer, ThreeDLayer):
    """Handles rendering a convolutional layer for a nn"""

    def __init__(
        self,
        num_feature_maps,
        feature_map_size=None,
        filter_size=None,
        cell_width=0.2,
        filter_spacing=0.1,
        color=BLUE,
        active_color=ORANGE,
        filter_color=ORANGE,
        show_grid_lines=False,
        fill_opacity=0.3,
        stride=1,
        stroke_width=2.0,
        activation_function=None,
        padding=0,
        padding_dashed=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_feature_maps = num_feature_maps
        self.filter_color = filter_color
        if isinstance(padding, tuple):
            assert len(padding) == 2
            self.padding = padding
        elif isinstance(padding, int):
            self.padding = (padding, padding)
        else:
            raise Exception(f"Unrecognized type for padding: {type(padding)}")

        if isinstance(feature_map_size, int):
            self.feature_map_size = (feature_map_size, feature_map_size)
        else:
            self.feature_map_size = feature_map_size

        if isinstance(filter_size, int):
            self.filter_size = (filter_size, filter_size)
        else:
            self.filter_size = filter_size

        self.cell_width = cell_width
        self.filter_spacing = filter_spacing
        self.color = color
        self.active_color = active_color
        self.stride = stride
        self.stroke_width = stroke_width
        self.show_grid_lines = show_grid_lines
        self.activation_function = activation_function
        self.fill_opacity = fill_opacity
        self.padding_dashed = padding_dashed

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs,
    ):
        # Make the feature maps
        self.feature_maps = self.construct_feature_maps()
        self.add(self.feature_maps)
        # Rotate stuff properly
        # normal_vector = self.feature_maps[0].get_normal_vector()
        self.rotate(
            ThreeDLayer.rotation_angle,
            about_point=self.get_center(),
            axis=ThreeDLayer.rotation_axis,
        )

        self.construct_activation_function()
        super().construct_layer(input_layer, output_layer, **kwargs)

    def construct_activation_function(self):
        """Construct the activation function"""
        # Add the activation function
        if not self.activation_function is None:
            # Check if it is a string
            if isinstance(self.activation_function, str):
                activation_function = get_activation_function_by_name(
                    self.activation_function
                )()
            else:
                assert isinstance(self.activation_function, ActivationFunction)
                activation_function = self.activation_function
            # Plot the function above the rest of the layer
            self.activation_function = activation_function
            self.add(self.activation_function)

    def construct_feature_maps(self):
        """Creates the neural network layer"""
        # Draw rectangles that are filled in with opacity
        feature_maps = []
        for filter_index in range(self.num_feature_maps):
            # Check if we need to add padding
            """
            feature_map = GriddedRectangle(
                color=self.color,
                height=self.feature_map_size[1] * self.cell_width,
                width=self.feature_map_size[0] * self.cell_width,
                fill_color=self.color,
                fill_opacity=self.fill_opacity,
                stroke_color=self.color,
                stroke_width=self.stroke_width,
                grid_xstep=self.cell_width,
                grid_ystep=self.cell_width,
                grid_stroke_width=self.stroke_width / 2,
                grid_stroke_color=self.color,
                show_grid_lines=self.show_grid_lines,
            )
            """
            # feature_map = GriddedRectangle()
            feature_map = FeatureMap(
                color=self.color,
                feature_map_size=self.feature_map_size,
                cell_width=self.cell_width,
                fill_color=self.color,
                fill_opacity=self.fill_opacity,
                padding=self.padding,
                padding_dashed=self.padding_dashed,
            )
            # Move the feature map
            feature_map.move_to([0, 0, filter_index * self.filter_spacing])
            # rectangle.set_z_index(4)
            feature_maps.append(feature_map)

        return VGroup(*feature_maps)

    def highlight_and_unhighlight_feature_maps(self):
        """Highlights then unhighlights feature maps"""
        return Succession(
            ApplyMethod(self.feature_maps.set_color, self.active_color),
            ApplyMethod(self.feature_maps.set_color, self.color),
        )

    def make_forward_pass_animation(self, run_time=5, layer_args={}, **kwargs):
        """Convolution forward pass animation"""
        # Note: most of this animation is done in the Convolution3DToConvolution3D layer
        if not self.activation_function is None:
            animation_group = AnimationGroup(
                self.activation_function.make_evaluate_animation(),
                self.highlight_and_unhighlight_feature_maps(),
                lag_ratio=0.0,
            )
        else:
            animation_group = AnimationGroup()

        return animation_group

    def scale(self, scale_factor, **kwargs):
        self.cell_width *= scale_factor
        super().scale(scale_factor, **kwargs)

    def get_center(self):
        """Overrides function for getting center

        The reason for this is so that the center calculation
        does not include the activation function.
        """
        return self.feature_maps.get_center()

    def get_width(self):
        """Overrides get width function"""
        return self.feature_maps.length_over_dim(0)

    def get_height(self):
        """Overrides get height function"""
        return self.feature_maps.length_over_dim(1)

    def move_to(self, mobject_or_point):
        """Moves the center of the layer to the given mobject or point"""
        layer_center = self.feature_maps.get_center()
        if isinstance(mobject_or_point, Mobject):
            target_center = mobject_or_point.get_center() 
        else:
            target_center = mobject_or_point

        self.shift(target_center - layer_center)

    @override_animation(Create)
    def _create_override(self, **kwargs):
        return FadeIn(self.feature_maps)



import random
from manim import *
from manim_ml.utils.mobjects.gridded_rectangle import GriddedRectangle
from manim_ml.neural_network.layers.convolutional_2d_to_convolutional_2d import (
    get_rotated_shift_vectors,
)

from manim_ml.neural_network.layers.max_pooling_2d import MaxPooling2DLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer, ThreeDLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer


class Uncreate(Create):
    def __init__(
        self,
        mobject,
        reverse_rate_function: bool = True,
        introducer: bool = True,
        remover: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            mobject,
            reverse_rate_function=reverse_rate_function,
            introducer=introducer,
            remover=remover,
            **kwargs,
        )


class Convolutional2DToMaxPooling2D(ConnectiveLayer, ThreeDLayer):
    """Feed Forward to Embedding Layer"""

    input_class = Convolutional2DLayer
    output_class = MaxPooling2DLayer

    def __init__(
        self,
        input_layer: Convolutional2DLayer,
        output_layer: MaxPooling2DLayer,
        active_color=ORANGE,
        **kwargs,
    ):
        super().__init__(input_layer, output_layer, **kwargs)
        self.active_color = active_color

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs,
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, layer_args={}, run_time=1.5, **kwargs):
        """Forward pass animation from conv2d to max pooling"""
        cell_width = self.input_layer.cell_width
        feature_map_size = self.input_layer.feature_map_size
        kernel_size = self.output_layer.kernel_size
        feature_maps = self.input_layer.feature_maps
        grid_stroke_width = 1.0
        # Make all of the kernel gridded rectangles
        create_gridded_rectangle_animations = []
        create_and_remove_cell_animations = []
        transform_gridded_rectangle_animations = []
        remove_gridded_rectangle_animations = []

        for feature_map_index, feature_map in enumerate(feature_maps):
            # 1. Draw gridded rectangle with kernel_size x kernel_size
            #   box regions over the input feature maps.
            gridded_rectangle = GriddedRectangle(
                color=self.active_color,
                height=cell_width * feature_map_size[1],
                width=cell_width * feature_map_size[0],
                grid_xstep=cell_width * kernel_size,
                grid_ystep=cell_width * kernel_size,
                grid_stroke_width=grid_stroke_width,
                grid_stroke_color=self.active_color,
                show_grid_lines=True,
            )
            gridded_rectangle.set_z_index(10)
            # 2. Randomly highlight one of the cells in the kernel.
            highlighted_cells = []
            num_cells_in_kernel = kernel_size * kernel_size
            num_x_kernels = int(feature_map_size[0] / kernel_size)
            num_y_kernels = int(feature_map_size[1] / kernel_size)
            for kernel_x in range(0, num_x_kernels):
                for kernel_y in range(0, num_y_kernels):
                    # Choose a random cell index
                    cell_index = random.randint(0, num_cells_in_kernel - 1)
                    # Make a rectangle in that cell
                    cell_rectangle = GriddedRectangle(
                        color=self.active_color,
                        height=cell_width,
                        width=cell_width,
                        fill_opacity=0.7,
                        stroke_width=0.0,
                        z_index=10,
                    )
                    # Move to the correct location
                    kernel_shift_vector = [
                        kernel_size * cell_width * kernel_x,
                        -1 * kernel_size * cell_width * kernel_y,
                        0,
                    ]
                    cell_shift_vector = [
                        (cell_index % kernel_size) * cell_width,
                        -1 * int(cell_index / kernel_size) * cell_width,
                        0,
                    ]
                    cell_rectangle.next_to(
                        gridded_rectangle.get_corners_dict()["top_left"],
                        submobject_to_align=cell_rectangle.get_corners_dict()[
                            "top_left"
                        ],
                        buff=0.0,
                    )
                    cell_rectangle.shift(kernel_shift_vector)
                    cell_rectangle.shift(cell_shift_vector)
                    highlighted_cells.append(cell_rectangle)
            # Rotate the gridded rectangles so they match the angle
            # of the conv maps
            gridded_rectangle_group = VGroup(gridded_rectangle, *highlighted_cells)
            gridded_rectangle_group.rotate(
                ThreeDLayer.rotation_angle,
                about_point=gridded_rectangle.get_center(),
                axis=ThreeDLayer.rotation_axis,
            )
            gridded_rectangle_group.next_to(
                feature_map.get_corners_dict()["top_left"],
                submobject_to_align=gridded_rectangle.get_corners_dict()["top_left"],
                buff=0.0,
            )
            # 3. Make a create gridded rectangle
            create_rectangle = Create(
                gridded_rectangle,
            )
            create_gridded_rectangle_animations.append(create_rectangle)
            # 4. Create and fade out highlighted cells
            create_group = AnimationGroup(
                *[Create(highlighted_cell) for highlighted_cell in highlighted_cells],
                lag_ratio=1.0,
            )
            uncreate_group = AnimationGroup(
                *[Uncreate(highlighted_cell) for highlighted_cell in highlighted_cells],
                lag_ratio=0.0,
            )
            create_and_remove_cell_animation = Succession(
                create_group, Wait(1.0), uncreate_group
            )
            create_and_remove_cell_animations.append(create_and_remove_cell_animation)
            # 5. Move and resize the gridded rectangle to the output
            #   feature maps.
            output_gridded_rectangle = GriddedRectangle(
                color=self.active_color,
                height=cell_width * feature_map_size[1] / 2,
                width=cell_width * feature_map_size[0] / 2,
                grid_xstep=cell_width,
                grid_ystep=cell_width,
                grid_stroke_width=grid_stroke_width,
                grid_stroke_color=self.active_color,
                show_grid_lines=True,
            )
            output_gridded_rectangle.rotate(
                ThreeDLayer.rotation_angle,
                about_point=output_gridded_rectangle.get_center(),
                axis=ThreeDLayer.rotation_axis,
            )
            output_gridded_rectangle.move_to(
                self.output_layer.feature_maps[feature_map_index].copy()
            )
            transform_rectangle = ReplacementTransform(
                gridded_rectangle,
                output_gridded_rectangle,
                introducer=True,
            )
            transform_gridded_rectangle_animations.append(
                transform_rectangle,
            )
            """
            Succession(
                Uncreate(gridded_rectangle),
                transform_rectangle,
                lag_ratio=1.0
            )
            """
            # 6. Make the gridded feature map(s) disappear.
            remove_gridded_rectangle_animations.append(
                Uncreate(gridded_rectangle_group)
            )

        create_gridded_rectangle_animation = AnimationGroup(
            *create_gridded_rectangle_animations
        )
        create_and_remove_cell_animation = AnimationGroup(
            *create_and_remove_cell_animations
        )
        transform_gridded_rectangle_animation = AnimationGroup(
            *transform_gridded_rectangle_animations
        )
        remove_gridded_rectangle_animation = AnimationGroup(
            *remove_gridded_rectangle_animations
        )

        return Succession(
            create_gridded_rectangle_animation,
            Wait(1),
            create_and_remove_cell_animation,
            transform_gridded_rectangle_animation,
            Wait(1),
            remove_gridded_rectangle_animation,
            lag_ratio=1.0,
        )



import manim.utils.opengl as opengl
from manim import *
from manim.opengl import *  # type: ignore


class InlineShaderExample(Scene):
    def construct(self):
        config["background_color"] = "#333333"

        c = Circle(fill_opacity=0.7).shift(UL)
        self.add(c)

        shader = Shader(
            self.renderer.context,
            source={
                "vertex_shader": """
                #version 330
                in vec4 in_vert;
                in vec4 in_color;
                out vec4 v_color;
                uniform mat4 u_model_view_matrix;
                uniform mat4 u_projection_matrix;
                void main() {
                    v_color = in_color;
                    vec4 camera_space_vertex = u_model_view_matrix * in_vert;
                    vec4 clip_space_vertex = u_projection_matrix * camera_space_vertex;
                    gl_Position = clip_space_vertex;
                }
            """,
                "fragment_shader": """
            #version 330
            in vec4 v_color;
            out vec4 frag_color;
            void main() {
              frag_color = v_color;
            }

            void main () {
                // Previously, you'd have rendered your complete scene into a texture
                // bound to "fullScreenTexture."
                vec4 rValue = texture2D(fullscreenTexture, gl_TexCoords[0] - rOffset);  
                vec4 gValue = texture2D(fullscreenTexture, gl_TexCoords[0] - gOffset);
                vec4 bValue = texture2D(fullscreenTexture, gl_TexCoords[0] - bOffset);  

                // Combine the offset colors.
                gl_FragColor = vec4(rValue.r, gValue.g, bValue.b, 1.0);
            }
            """,
            },
        )
        shader.set_uniform("u_model_view_matrix", opengl.view_matrix())
        shader.set_uniform(
            "u_projection_matrix",
            opengl.orthographic_projection_matrix(),
        )

        attributes = np.zeros(
            6,
            dtype=[
                ("in_vert", np.float32, (4,)),
                ("in_color", np.float32, (4,)),
            ],
        )
        attributes["in_vert"] = np.array(
            [
                [-1, -1, 0, 1],
                [-1, 1, 0, 1],
                [1, 1, 0, 1],
                [-1, -1, 0, 1],
                [1, -1, 0, 1],
                [1, 1, 0, 1],
            ],
        )
        attributes["in_color"] = np.array(
            [
                [0, 0, 1, 1],
                [0, 0, 1, 1],
                [0, 0, 1, 1],
                [0, 0, 1, 1],
                [0, 0, 1, 1],
                [0, 0, 1, 1],
            ],
        )
        mesh = Mesh(shader, attributes)
        self.add(mesh)

        self.wait(5)
        # self.embed_2()



from manim import *


class ManimML3DScene(ThreeDScene):
    """
    This is a wrapper class for the Manim ThreeDScene

    Note: the primary purpose of this is to make it so
    that everything inside of a layer
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def play(self):
        """ """
        pass



import warnings

from manim import *
from manim_ml.neural_network.layers.parent_layers import BlankConnective, ThreeDLayer
from manim_ml.neural_network.layers import connective_layers_list


def get_connective_layer(input_layer, output_layer):
    """
    Deduces the relevant connective layer
    """
    connective_layer_class = None
    for candidate_class in connective_layers_list:
        input_class = candidate_class.input_class
        output_class = candidate_class.output_class
        if isinstance(input_layer, input_class) and isinstance(
            output_layer, output_class
        ):
            connective_layer_class = candidate_class
            break

    if connective_layer_class is None:
        connective_layer_class = BlankConnective
        warnings.warn(
            f"Unrecognized input/output class pair: {input_layer} and {output_layer}"
        )
    # Make the instance now
    connective_layer = connective_layer_class(input_layer, output_layer)

    return connective_layer



from manim import *
from PIL import Image
from manim_ml.neural_network.layers import EmbeddingLayer, FeedForwardLayer, ImageLayer
from manim_ml.neural_network.neural_network import NeuralNetwork

config.pixel_height = 720
config.pixel_width = 1280
config.frame_height = 8.0
config.frame_width = 8.0


class VariationalAutoencoderScene(Scene):
    def construct(self):
        embedding_layer = EmbeddingLayer(dist_theme="ellipse").scale(2)

        image = Image.open("images/image.jpeg")
        numpy_image = np.asarray(image)
        # Make nn
        neural_network = NeuralNetwork(
            [
                ImageLayer(numpy_image, height=1.4),
                FeedForwardLayer(5),
                FeedForwardLayer(3),
                embedding_layer,
                FeedForwardLayer(3),
                FeedForwardLayer(5),
                ImageLayer(numpy_image, height=1.4),
            ],
            layer_spacing=0.1,
        )

        neural_network.scale(1.3)

        self.play(Create(neural_network), run_time=3)
        self.play(neural_network.make_forward_pass_animation(), run_time=5)
        self.play(neural_network.make_forward_pass_animation(), run_time=5)



from manim import *
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.feed_forward_to_feed_forward import (
    FeedForwardToFeedForward,
)
from manim_ml.neural_network.layers.util import get_connective_layer


def test_get_connective_layer():
    """Tests get connective layer"""
    input_layer = FeedForwardLayer(3)
    output_layer = FeedForwardLayer(5)
    connective_layer = get_connective_layer(input_layer, output_layer)

    assert isinstance(connective_layer, FeedForwardToFeedForward)



from manim_ml.neural_network.layers.convolutional_2d_to_feed_forward import (
    Convolutional2DToFeedForward,
)
from manim_ml.neural_network.layers.convolutional_2d_to_max_pooling_2d import (
    Convolutional2DToMaxPooling2D,
)
from manim_ml.neural_network.layers.image_to_convolutional_2d import (
    ImageToConvolutional2DLayer,
)
from manim_ml.neural_network.layers.max_pooling_2d_to_convolutional_2d import (
    MaxPooling2DToConvolutional2D,
)
from manim_ml.neural_network.layers.max_pooling_2d_to_feed_forward import (
    MaxPooling2DToFeedForward,
)
from .convolutional_2d_to_convolutional_2d import Convolutional2DToConvolutional2D
from .convolutional_2d import Convolutional2DLayer
from .feed_forward_to_vector import FeedForwardToVector
from .paired_query_to_feed_forward import PairedQueryToFeedForward
from .embedding_to_feed_forward import EmbeddingToFeedForward
from .embedding import EmbeddingLayer
from .feed_forward_to_embedding import FeedForwardToEmbedding
from .feed_forward_to_feed_forward import FeedForwardToFeedForward
from .feed_forward_to_image import FeedForwardToImage
from .feed_forward import FeedForwardLayer
from .image_to_feed_forward import ImageToFeedForward
from .image import ImageLayer
from .parent_layers import ConnectiveLayer, NeuralNetworkLayer
from .triplet import TripletLayer
from .triplet_to_feed_forward import TripletToFeedForward
from .paired_query import PairedQueryLayer
from .paired_query_to_feed_forward import PairedQueryToFeedForward
from .max_pooling_2d import MaxPooling2DLayer
from .feed_forward_to_math_operation import FeedForwardToMathOperation

connective_layers_list = (
    EmbeddingToFeedForward,
    FeedForwardToEmbedding,
    FeedForwardToFeedForward,
    FeedForwardToImage,
    ImageToFeedForward,
    PairedQueryToFeedForward,
    TripletToFeedForward,
    PairedQueryToFeedForward,
    FeedForwardToVector,
    Convolutional2DToConvolutional2D,
    ImageToConvolutional2DLayer,
    Convolutional2DToFeedForward,
    Convolutional2DToMaxPooling2D,
    MaxPooling2DToConvolutional2D,
    MaxPooling2DToFeedForward,
    FeedForwardToMathOperation
)



from typing import List, Union
import numpy as np

from manim import *
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer
import manim_ml

class FeedForwardToFeedForward(ConnectiveLayer):
    """Layer for connecting FeedForward layer to FeedForwardLayer"""

    input_class = FeedForwardLayer
    output_class = FeedForwardLayer

    def __init__(
        self,
        input_layer,
        output_layer,
        passing_flash=True,
        dot_radius=0.05,
        animation_dot_color=manim_ml.config.color_scheme.active_color,
        edge_color=manim_ml.config.color_scheme.secondary_color,
        edge_width=1.5,
        camera=None,
        **kwargs
    ):
        super().__init__(input_layer, output_layer, **kwargs)
        self.passing_flash = passing_flash
        self.edge_color = edge_color
        self.dot_radius = dot_radius
        self.animation_dot_color = animation_dot_color
        self.edge_width = edge_width

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        self.edges = self.construct_edges()
        self.add(self.edges)
        super().construct_layer(input_layer, output_layer, **kwargs)

    def construct_edges(self):
        # Go through each node in the two layers and make a connecting line
        edges = []
        for node_i in self.input_layer.node_group:
            for node_j in self.output_layer.node_group:
                line = Line(
                    node_i.get_center(),
                    node_j.get_center(),
                    color=self.edge_color,
                    stroke_width=self.edge_width,
                )
                edges.append(line)

        edges = VGroup(*edges)
        return edges

    @override_animation(FadeOut)
    def _fadeout_animation(self):
        animations = []

        for edge in self.edges:
            animations.append(FadeOut(edge))

        animation_group = AnimationGroup(*animations)

        return animation_group

    def make_forward_pass_animation(
        self, layer_args={}, run_time=1, feed_forward_dropout=0.0, **kwargs
    ):
        """Animation for passing information from one FeedForwardLayer to the next"""
        path_animations = []
        dots = []
        for edge_index, edge in enumerate(self.edges):
            if (
                not "edge_indices_to_dropout" in layer_args
                or not edge_index in layer_args["edge_indices_to_dropout"]
            ):
                dot = Dot(
                    color=self.animation_dot_color,
                    fill_opacity=1.0,
                    radius=self.dot_radius,
                )
                # Add to dots group
                dots.append(dot)
                # Make the animation
                if self.passing_flash:
                    copy_edge = edge.copy()
                    anim = ShowPassingFlash(
                        copy_edge.set_color(self.animation_dot_color), time_width=0.3
                    )
                else:
                    anim = MoveAlongPath(
                        dot, edge, run_time=run_time, rate_function=sigmoid
                    )
                path_animations.append(anim)

        if not self.passing_flash:
            dots = VGroup(*dots)
            self.add(dots)

        path_animations = AnimationGroup(*path_animations)

        return path_animations

    def modify_edge_colors(self, colors=None, magnitudes=None, color_scheme="inferno"):
        """Changes the colors of edges"""
        # TODO implement
        pass

    def modify_edge_stroke_widths(self, widths):
        """Changes the widths of the edges"""
        assert len(widths) > 0
        # Note: 1d-arrays are assumed to be in row major order
        widths = np.array(widths)
        widths = np.flatten(widths)
        # Check thickness size
        assert np.shape(widths)[0] == len(self.edges)
        # Make animation
        animations = []

        for index, edge in enumerate(self.edges):
            width = widths[index]
            change_width = edge.animate.set_stroke_width(width)
            animations.append(change_width)

        animation_group = AnimationGroup(*animations)

        return animation_group

    @override_animation(Create)
    def _create_override(self, **kwargs):
        animations = []

        for edge in self.edges:
            animations.append(Create(edge))

        animation_group = AnimationGroup(*animations, lag_ratio=0.0)
        return animation_group



from manim import *
import numpy as np


class GriddedRectangle(VGroup):
    """Rectangle object with grid lines"""

    def __init__(
        self,
        color=ORANGE,
        height=2.0,
        width=4.0,
        mark_paths_closed=True,
        close_new_points=True,
        grid_xstep=None,
        grid_ystep=None,
        grid_stroke_width=0.0,  # DEFAULT_STROKE_WIDTH/2,
        grid_stroke_color=ORANGE,
        grid_stroke_opacity=1.0,
        stroke_width=2.0,
        fill_opacity=0.2,
        show_grid_lines=False,
        dotted_lines=False,
        **kwargs
    ):
        super().__init__()
        # Fields
        self.color = color
        self.mark_paths_closed = mark_paths_closed
        self.close_new_points = close_new_points
        self.grid_xstep = grid_xstep
        self.grid_ystep = grid_ystep
        self.grid_stroke_width = grid_stroke_width
        self.grid_stroke_color = grid_stroke_color
        self.grid_stroke_opacity = grid_stroke_opacity if show_grid_lines else 0.0
        self.stroke_width = stroke_width
        self.rotation_angles = [0, 0, 0]
        self.show_grid_lines = show_grid_lines
        self.untransformed_width = width
        self.untransformed_height = height
        self.dotted_lines = dotted_lines
        # Make rectangle
        if self.dotted_lines:
            no_border_rectangle = Rectangle(
                width=width,
                height=height,
                color=color,
                fill_color=color,
                stroke_opacity=0.0,
                fill_opacity=fill_opacity,
                shade_in_3d=True,
            )
            self.rectangle = no_border_rectangle
            border_rectangle = Rectangle(
                width=width,
                height=height,
                color=color,
                fill_color=color,
                fill_opacity=fill_opacity,
                shade_in_3d=True,
                stroke_width=stroke_width,
            )
            self.dotted_lines = DashedVMobject(
                border_rectangle,
                num_dashes=int((width + height) / 2) * 20,
            )
            self.add(self.dotted_lines)
        else:
            self.rectangle = Rectangle(
                width=width,
                height=height,
                color=color,
                stroke_width=stroke_width,
                fill_color=color,
                fill_opacity=fill_opacity,
                shade_in_3d=True,
            )
        self.add(self.rectangle)
        # Make grid lines
        grid_lines = self.make_grid_lines()
        self.add(grid_lines)
        # Make corner rectangles
        self.corners_dict = self.make_corners_dict()
        self.add(*self.corners_dict.values())

    def make_corners_dict(self):
        """Make corners dictionary"""
        corners_dict = {
            "top_right": Dot(
                self.rectangle.get_corner([1, 1, 0]), fill_opacity=0.0, radius=0.0
            ),
            "top_left": Dot(
                self.rectangle.get_corner([-1, 1, 0]), fill_opacity=0.0, radius=0.0
            ),
            "bottom_left": Dot(
                self.rectangle.get_corner([-1, -1, 0]), fill_opacity=0.0, radius=0.0
            ),
            "bottom_right": Dot(
                self.rectangle.get_corner([1, -1, 0]), fill_opacity=0.0, radius=0.0
            ),
        }

        return corners_dict

    def get_corners_dict(self):
        """Returns a dictionary of the corners"""
        # Sort points through clockwise rotation of a vector in the xy plane
        return self.corners_dict

    def make_grid_lines(self):
        """Make grid lines in rectangle"""
        grid_lines = VGroup()

        v = self.rectangle.get_vertices()
        if self.grid_xstep is not None:
            grid_xstep = abs(self.grid_xstep)
            count = int(self.width / grid_xstep)
            grid = VGroup(
                *(
                    Line(
                        v[1] + i * grid_xstep * RIGHT,
                        v[1] + i * grid_xstep * RIGHT + self.height * DOWN,
                        stroke_color=self.grid_stroke_color,
                        stroke_width=self.grid_stroke_width,
                        stroke_opacity=self.grid_stroke_opacity,
                        shade_in_3d=True,
                    )
                    for i in range(1, count)
                )
            )
            grid_lines.add(grid)

        if self.grid_ystep is not None:
            grid_ystep = abs(self.grid_ystep)
            count = int(self.height / grid_ystep)
            grid = VGroup(
                *(
                    Line(
                        v[1] + i * grid_ystep * DOWN,
                        v[1] + i * grid_ystep * DOWN + self.width * RIGHT,
                        stroke_color=self.grid_stroke_color,
                        stroke_width=self.grid_stroke_width,
                        stroke_opacity=self.grid_stroke_opacity,
                    )
                    for i in range(1, count)
                )
            )
            grid_lines.add(grid)

        return grid_lines

    def get_center(self):
        return self.rectangle.get_center()

    def get_normal_vector(self):
        vertex_1 = self.rectangle.get_vertices()[0]
        vertex_2 = self.rectangle.get_vertices()[1]
        vertex_3 = self.rectangle.get_vertices()[2]
        # First vector
        normal_vector = np.cross((vertex_1 - vertex_2), (vertex_1 - vertex_3))

        return normal_vector

    def set_color(self, color):
        """Sets the color of the gridded rectangle"""
        self.color = color
        self.rectangle.set_color(color)
        self.rectangle.set_stroke_color(color)



from manim import *
from dataclasses import dataclass

@dataclass
class ColorScheme:
    primary_color: str
    secondary_color: str
    active_color: str
    text_color: str
    background_color: str

dark_mode = ColorScheme(
    primary_color=BLUE,
    secondary_color=WHITE,
    active_color=ORANGE,
    text_color=WHITE,
    background_color=BLACK
)

light_mode = ColorScheme(
    primary_color=BLUE,
    secondary_color=BLACK,
    active_color=ORANGE,
    text_color=BLACK,
    background_color=WHITE
)



import manim_ml
from manim_ml.neural_network.neural_network import NeuralNetwork
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer


class FeedForwardNeuralNetwork(NeuralNetwork):
    """NeuralNetwork with just feed forward layers"""

    def __init__(
        self, 
        layer_node_count, 
        node_radius=0.08, 
        node_color=manim_ml.config.color_scheme.primary_color, 
        **kwargs
    ):
        # construct layer
        layers = []
        for num_nodes in layer_node_count:
            layer = FeedForwardLayer(
                num_nodes, node_color=node_color, node_radius=node_radius
            )
            layers.append(layer)
        # call super class
        super().__init__(layers, **kwargs)



"""
    Transformations for manipulating a neural network object. 
"""
from manim import *
from manim_ml.neural_network.layers.util import get_connective_layer


class RemoveLayer(AnimationGroup):
    """
    Animation for removing a layer from a neural network.

    Note: I needed to do something strange for creating the new connective layer.
    The issue with creating it initially is that the positions of the sides of the
    connective layer depend upon the location of the moved layers **after** the
    move animations are performed. However, all of these animations are performed
    after the animations have been created. This means that the animation depends upon
    the state of the neural network layers after previous animations have been run.
    To fix this issue I needed to use an UpdateFromFunc.
    """

    def __init__(self, layer, neural_network, layer_spacing=0.2):
        self.layer = layer
        self.neural_network = neural_network
        self.layer_spacing = layer_spacing
        # Get the before and after layers
        layers_tuple = self.get_connective_layers()
        self.before_layer = layers_tuple[0]
        self.after_layer = layers_tuple[1]
        self.before_connective = layers_tuple[2]
        self.after_connective = layers_tuple[3]
        # Make the animations
        remove_animations = self.make_remove_animation()
        move_animations = self.make_move_animation()
        new_connective_animation = self.make_new_connective_animation()
        # Add all of the animations to the group
        animations_list = [remove_animations, move_animations, new_connective_animation]

        super().__init__(*animations_list, lag_ratio=1.0)

    def get_connective_layers(self):
        """Gets the connective layers before and after self.layer"""
        # Get layer index
        layer_index = self.neural_network.all_layers.index_of(self.layer)
        if layer_index == -1:
            raise Exception("Layer object not found")
        # Get the layers before and after
        before_layer = None
        after_layer = None
        before_connective = None
        after_connective = None
        if layer_index - 2 >= 0:
            before_layer = self.neural_network.all_layers[layer_index - 2]
            before_connective = self.neural_network.all_layers[layer_index - 1]
        if layer_index + 2 < len(self.neural_network.all_layers):
            after_layer = self.neural_network.all_layers[layer_index + 2]
            after_connective = self.neural_network.all_layers[layer_index + 1]

        return before_layer, after_layer, before_connective, after_connective

    def make_remove_animation(self):
        """Removes layer and the surrounding connective layers"""
        remove_layer_animation = self.make_remove_layer_animation()
        remove_connective_animation = self.make_remove_connective_layers_animation()
        # Remove animations
        remove_animations = AnimationGroup(
            remove_layer_animation, remove_connective_animation
        )

        return remove_animations

    def make_remove_layer_animation(self):
        """Removes the layer"""
        # Remove the layer
        self.neural_network.all_layers.remove(self.layer)
        # Fade out the removed layer
        fade_out_removed = FadeOut(self.layer)
        return fade_out_removed

    def make_remove_connective_layers_animation(self):
        """Removes the connective layers before and after layer if they exist"""
        # Fade out the removed connective layers
        fade_out_before_connective = AnimationGroup()
        if not self.before_connective is None:
            self.neural_network.all_layers.remove(self.before_connective)
            fade_out_before_connective = FadeOut(self.before_connective)
        fade_out_after_connective = AnimationGroup()
        if not self.after_connective is None:
            self.neural_network.all_layers.remove(self.after_connective)
            fade_out_after_connective = FadeOut(self.after_connective)
        # Group items
        remove_connective_group = AnimationGroup(
            fade_out_after_connective, fade_out_before_connective
        )

        return remove_connective_group

    def make_move_animation(self):
        """Collapses layers"""
        # Animate the movements
        move_before_layers = AnimationGroup()
        shift_right_amount = None
        if not self.before_layer is None:
            # Compute shift amount
            layer_dist = np.abs(
                self.layer.get_center() - self.before_layer.get_right()
            )[0]
            shift_right_amount = np.array([layer_dist - self.layer_spacing / 2, 0, 0])
            # Shift all layers before forward
            before_layer_index = self.neural_network.all_layers.index_of(
                self.before_layer
            )
            layers_before = Group(
                *self.neural_network.all_layers[: before_layer_index + 1]
            )
            move_before_layers = layers_before.animate.shift(shift_right_amount)
        move_after_layers = AnimationGroup()
        shift_left_amount = None
        if not self.after_layer is None:
            layer_dist = np.abs(self.after_layer.get_left() - self.layer.get_center())[
                0
            ]
            shift_left_amount = np.array([-layer_dist + self.layer_spacing / 2, 0, 0])
            # Shift all layers after backward
            after_layer_index = self.neural_network.all_layers.index_of(
                self.after_layer
            )
            layers_after = Group(*self.neural_network.all_layers[after_layer_index:])
            move_after_layers = layers_after.animate.shift(shift_left_amount)
        # Group the move animations
        move_group = AnimationGroup(move_before_layers, move_after_layers)

        return move_group

    def make_new_connective_animation(self):
        """Makes new connective layer"""
        self.anim_count = 0

        def create_new_connective(neural_network):
            """
            Creates new connective layer

            This is a closure that creates a new connective layer and animates it.
            """
            self.anim_count += 1
            if self.anim_count == 1:
                if not self.before_layer is None and not self.after_layer is None:
                    print(neural_network)
                    new_connective_class = get_connective_layer(
                        self.before_layer, self.after_layer
                    )
                    before_layer_index = (
                        neural_network.all_layers.index_of(self.before_layer) + 1
                    )
                    neural_network.all_layers.insert(before_layer_index, new_connective)
                    print(neural_network)

        update_func_anim = UpdateFromFunc(self.neural_network, create_new_connective)

        return update_func_anim


class InsertLayer(AnimationGroup):
    """Animation for inserting layer at given index"""

    def __init__(self, layer, index, neural_network):
        self.layer = layer
        self.index = index
        self.neural_network = neural_network
        # Check valid index
        assert index < len(self.neural_network.all_layers)
        # Layers before and after
        self.layers_before = self.neural_network.all_layers[: self.index]
        self.layers_after = self.neural_network.all_layers[self.index :]
        # Get the non-connective layer before and after
        if len(self.layers_before) > 0:
            self.layer_before = self.layers_before[-2]
        if len(self.layers_after) > 0:
            self.layer_after = self.layers_after[0]
        # Move layer
        if not self.layer_after is None:
            self.layer.move_to(self.layer_after)
        # Make animations
        (
            self.old_connective_layer,
            remove_connective_layer,
        ) = self.remove_connective_layer_animation()
        move_layers = self.make_move_layers_animation()
        create_layer = self.make_create_layer_animation()
        # create_connective_layers = self.make_create_connective_layers()
        animations = [
            remove_connective_layer,
            move_layers,
            create_layer,
            #    create_connective_layers
        ]

        super().__init__(*animations, lag_ratio=1.0)

    def get_connective_layer_widths(self):
        """Gets the widths of the connective layers"""
        # Make the layers
        before_connective = None
        after_connective = None
        # Get the connective layer objects
        if len(self.layers_before) > 0:
            before_connective = get_connective_layer(self.layer_before, self.layer)
        if len(self.layers_after) > 0:
            after_connective = get_connective_layer(self.layer, self.layer_after)
        # Compute the widths
        before_connective_width = 0
        if not before_connective is None:
            before_connective_width = before_connective.width
        after_connective_width = 0
        if not after_connective is None:
            after_connective_width = after_connective.width
        return before_connective_width, after_connective_width

    def remove_connective_layer_animation(self):
        """Removes the connective layer before the insertion index"""
        # Check if connective layer before exists
        if len(self.layers_before) > 0:
            removed_connective = self.layers_before[-1]
            self.layers_before.remove(removed_connective)
            self.neural_network.all_layers.remove(removed_connective)
            # Make remove animation
            remove_animation = FadeOut(removed_connective)
            return removed_connective, remove_animation

        return None, AnimationGroup()

    def make_move_layers_animation(self):
        """Shifts layers before and after"""
        (
            before_connective_width,
            after_connective_width,
        ) = self.get_connective_layer_widths()
        old_connective_width = 0
        if not self.old_connective_layer is None:
            old_connective_width = self.old_connective_layer.width
        # Before layer shift
        before_shift_animation = AnimationGroup()
        if len(self.layers_before) > 0:
            before_shift = np.array(
                [
                    -self.layer.width / 2
                    - before_connective_width
                    + old_connective_width,
                    0,
                    0,
                ]
            )
            # Shift layers before
            before_shift_animation = Group(*self.layers_before).animate.shift(
                before_shift
            )
        # After layer shift
        after_shift_animation = AnimationGroup()
        if len(self.layers_after) > 0:
            after_shift = np.array(
                [self.layer.width / 2 + after_connective_width, 0, 0]
            )
            # Shift layers after
            after_shift_animation = Group(*self.layers_after).animate.shift(after_shift)
        # Make animation group
        shift_animations = AnimationGroup(before_shift_animation, after_shift_animation)

        return shift_animations

    def make_create_layer_animation(self):
        """Animates the creation of the layer"""
        return Create(self.layer)

    def make_create_connective_layers_animation(
        self, before_connective, after_connective
    ):
        """Create connective layers"""
        # Make the layers
        before_connective = None
        after_connective = None
        # Get the connective layer objects
        if len(self.layers_before) > 0:
            before_connective = get_connective_layer(self.layers_before[-1], self.layer)
        if len(self.layers_after) > 0:
            after_connective = get_connective_layer(self.layers_after[0], self.layer)
        # Insert the layers
        # Make the animation
        animation_group = AnimationGroup(
            Create(before_connective), Create(after_connective)
        )

        return animation_group



import numpy as np
from manim import *


class NetworkConnection(VGroup):
    """
    This class allows for creating connections
    between locations in a network
    """

    direction_vector_map = {"up": UP, "down": DOWN, "left": LEFT, "right": RIGHT}

    def __init__(
        self,
        start_mobject,
        end_mobject,
        arc_direction="straight",
        buffer=0.0,
        arc_distance=0.2,
        stroke_width=2.0,
        color=WHITE,
        active_color=ORANGE,
    ):
        """Creates an arrow with right angles in it connecting
        two mobjects.

        Parameters
        ----------
        start_mobject : Mobject
            Mobject where the start of the connection is from
        end_mobject : Mobject
            Mobject where the end of the connection goes to
        arc_direction : str, optional
            direction that the connection arcs, by default "straight"
        buffer : float, optional
            amount of space between the connection and mobjects at the end
        arc_distance : float, optional
            Distance from start and end mobject that the arc bends
        stroke_width : float, optional
            Stroke width of the connection
        color : [float], optional
            Color of the connection
        active_color : [float], optional
            Color of active animations for this mobject
        """
        super().__init__()
        assert arc_direction in ["straight", "up", "down", "left", "right"]
        self.start_mobject = start_mobject
        self.end_mobject = end_mobject
        self.arc_direction = arc_direction
        self.buffer = buffer
        self.arc_distance = arc_distance
        self.stroke_width = stroke_width
        self.color = color
        self.active_color = active_color

        self.make_mobjects()

    def make_mobjects(self):
        """Makes the submobjects"""
        if self.start_mobject.get_center()[0] < self.end_mobject.get_center()[0]:
            left_mobject = self.start_mobject
            right_mobject = self.end_mobject
        else:
            right_mobject = self.start_mobject
            left_mobject = self.end_mobject
        if self.arc_direction == "straight":
            # Make an arrow
            self.straight_arrow = Arrow(
                start=left_mobject.get_right() + np.array([self.buffer, 0.0, 0.0]),
                end=right_mobject.get_left() + np.array([-1 * self.buffer, 0.0, 0.0]),
                color=WHITE,
                fill_color=WHITE,
                stroke_opacity=1.0,
                buff=0.0,
            )
            self.add(self.straight_arrow)
        else:
            # Figure out the direction of the arc
            direction_vector = NetworkConnection.direction_vector_map[
                self.arc_direction
            ]
            # Based on the position of the start and end layer, and direction
            # figure out how large to make each line
            # Whichever mobject has a critical point the farthest
            # distance in the direction_vector direction we will use that end
            left_mobject_critical_point = left_mobject.get_critical_point(direction_vector)
            right_mobject_critical_point = right_mobject.get_critical_point(direction_vector)
            # Take the dot product of each
            # These dot products correspond to the orthogonal projection
            # onto the direction vectors
            left_dot_product = np.dot(
                left_mobject_critical_point, 
                direction_vector
            )
            right_dot_product = np.dot(
                right_mobject_critical_point, 
                direction_vector
            )
            extra_distance = abs(left_dot_product - right_dot_product)
            # The difference between the dot products 
            if left_dot_product < right_dot_product:
                right_is_farthest = False
            else:
                right_is_farthest = True
            # Make the start arc piece
            start_line_start = left_mobject.get_critical_point(direction_vector)
            start_line_start += direction_vector * self.buffer
            start_line_end = start_line_start + direction_vector * self.arc_distance
            if not right_is_farthest:
                start_line_end = start_line_end + direction_vector * extra_distance
            self.start_line = Line(
                start_line_start,
                start_line_end,
                color=self.color,
                stroke_width=self.stroke_width,
            )
            # Make the end arc piece with an arrow
            end_line_end = right_mobject.get_critical_point(direction_vector)
            end_line_end += direction_vector * self.buffer
            end_line_start = end_line_end + direction_vector * self.arc_distance
            if right_is_farthest:
                end_line_start = end_line_start + direction_vector * extra_distance

            self.end_arrow = Arrow(
                start=end_line_start,
                end=end_line_end,
                color=WHITE,
                fill_color=WHITE,
                stroke_opacity=1.0,
                buff=0.0,
            )
            # Make the middle arc piece
            self.middle_line = Line(
                start_line_end,
                end_line_start,
                color=self.color,
                stroke_width=self.stroke_width,
            )
            # Add the mobjects
            self.add(
                self.start_line,
                self.middle_line,
                self.end_arrow,
            )

    @override_animation(ShowPassingFlash)
    def _override_passing_flash(self, run_time=1.0, time_width=0.2):
        """Passing flash animation"""
        if self.arc_direction == "straight":
            return ShowPassingFlash(
                self.straight_arrow.copy().set_color(self.active_color),
                time_width=time_width,
            )
        else:
            # Animate the start line
            start_line_animation = ShowPassingFlash(
                self.start_line.copy().set_color(self.active_color),
                time_width=time_width,
            )
            # Animate the middle line
            middle_line_animation = ShowPassingFlash(
                self.middle_line.copy().set_color(self.active_color),
                time_width=time_width,
            )
            # Animate the end line
            end_line_animation = ShowPassingFlash(
                self.end_arrow.copy().set_color(self.active_color),
                time_width=time_width,
            )

            return AnimationGroup(
                start_line_animation,
                middle_line_animation,
                end_line_animation,
                lag_ratio=1.0,
                run_time=run_time,
            )



from manim import *
from manim_ml.diffusion.mcmc import (
    MCMCAxes,
    MultidimensionalGaussianPosterior,
    metropolis_hastings_sampler,
)
from manim_ml.utils.mobjects.plotting import convert_matplotlib_figure_to_image_mobject

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib
plt.style.use('dark_background')

# Make the specific scene
config.pixel_height = 1200
config.pixel_width = 1200
config.frame_height = 7.0
config.frame_width = 7.0

def test_metropolis_hastings_sampler(iterations=100):
    samples, _, candidates = metropolis_hastings_sampler(iterations=iterations)
    assert samples.shape == (iterations, 2)

def plot_hexbin_gaussian_on_image_mobject(
    sample_func, 
    xlim=(-4, 4),
    ylim=(-4, 4)
):
    # Fixing random state for reproducibility
    np.random.seed(19680801)
    n = 100_000
    samples = []
    for i in range(n):
        samples.append(sample_func())
    samples = np.array(samples)

    x = samples[:, 0]
    y = samples[:, 1]
    
    fig, ax0 = plt.subplots(1, figsize=(5, 5))

    hb = ax0.hexbin(x, y, gridsize=50, cmap='gist_heat')

    ax0.set(xlim=xlim, ylim=ylim)

    return convert_matplotlib_figure_to_image_mobject(fig)

class MCMCTest(Scene):

    def construct(
        self, 
        mu=np.array([0.0, 0.0]), 
        var=np.array([[1.0, 1.0]])
    ):

        def gaussian_sample_func():
            vals = np.random.multivariate_normal(
                mu, 
                np.eye(2) * var, 
                1
            )[0]

            return vals

        image_mobject = plot_hexbin_gaussian_on_image_mobject(
            gaussian_sample_func
        )
        self.add(image_mobject)
        self.play(FadeOut(image_mobject))

        axes = MCMCAxes(
            x_range=[-4, 4],
            y_range=[-4, 4],
        )
        self.play(
            Create(axes)
        )

        gaussian_posterior = MultidimensionalGaussianPosterior(
            mu=np.array([0.0, 0.0]), 
            var=np.array([1.0, 1.0])
        )

        chain_sampling_animation, lines = axes.visualize_metropolis_hastings_chain_sampling(
            log_prob_fn=gaussian_posterior, 
            sampling_kwargs={"iterations": 500},
        )

        self.play(
            chain_sampling_animation,
            run_time=3.5
        )
        self.play(
            FadeOut(lines)
        )
        self.wait(1)
        self.play(
            FadeIn(image_mobject)
        )





from manim import *
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.image import ImageLayer
from manim_ml.neural_network.neural_network import NeuralNetwork
from PIL import Image
from manim_ml.utils.mobjects.image import GrayscaleImageMobject
import numpy as np

# Make the specific scene
config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 6.0
config.frame_width = 6.0

class CustomCamera(ThreeDCamera):

    def __init__(
        self,
        frame=None,
        fixed_dimension=0,
        default_frame_stroke_color=WHITE,
        default_frame_stroke_width=0,
        **kwargs,
    ):
        """
        Frame is a Mobject, (should almost certainly be a rectangle)
        determining which region of space the camera displays
        """
        self.fixed_dimension = fixed_dimension
        self.default_frame_stroke_color = default_frame_stroke_color
        self.default_frame_stroke_width = default_frame_stroke_width
        if frame is None:
            frame = ScreenRectangle(height=config["frame_height"])
            frame.set_stroke(
                self.default_frame_stroke_color,
                self.default_frame_stroke_width,
            )
        self.frame = frame
        super().__init__(**kwargs)

    # TODO, make these work for a rotated frame
    @property
    def frame_height(self):
        """Returns the height of the frame.

        Returns
        -------
        float
            The height of the frame.
        """
        return self.frame.height

    @property
    def frame_width(self):
        """Returns the width of the frame

        Returns
        -------
        float
            The width of the frame.
        """
        return self.frame.width

    @property
    def frame_center(self):
        """Returns the centerpoint of the frame in cartesian coordinates.

        Returns
        -------
        np.array
            The cartesian coordinates of the center of the frame.
        """
        return self.frame.get_center()

    @frame_height.setter
    def frame_height(self, frame_height: float):
        """Sets the height of the frame in MUnits.

        Parameters
        ----------
        frame_height
            The new frame_height.
        """
        self.frame.stretch_to_fit_height(frame_height)

    @frame_width.setter
    def frame_width(self, frame_width: float):
        """Sets the width of the frame in MUnits.

        Parameters
        ----------
        frame_width
            The new frame_width.
        """
        self.frame.stretch_to_fit_width(frame_width)

    @frame_center.setter
    def frame_center(self, frame_center: np.ndarray | list | tuple | Mobject):
        """Sets the centerpoint of the frame.

        Parameters
        ----------
        frame_center
            The point to which the frame must be moved.
            If is of type mobject, the frame will be moved to
            the center of that mobject.
        """
        self.frame.move_to(frame_center)

    def capture_mobjects(self, mobjects, **kwargs):
        # self.reset_frame_center()
        # self.realign_frame_shape()
        super().capture_mobjects(mobjects, **kwargs)


    # Since the frame can be moving around, the cairo
    # context used for updating should be regenerated
    # at each frame.  So no caching.
    def get_cached_cairo_context(self, pixel_array):
        """
        Since the frame can be moving around, the cairo
        context used for updating should be regenerated
        at each frame.  So no caching.
        """
        return None


    def cache_cairo_context(self, pixel_array, ctx):
        """
        Since the frame can be moving around, the cairo
        context used for updating should be regenerated
        at each frame.  So no caching.
        """
        pass

    def get_mobjects_indicating_movement(self):
        """
        Returns all mobjects whose movement implies that the camera
        should think of all other mobjects on the screen as moving

        Returns
        -------
        list
        """
        return [self.frame]

    def auto_zoom(
        self,
        mobjects: list[Mobject],
        margin: float = 0,
        only_mobjects_in_frame: bool = False,
        animate: bool = True,
    ):
        """Zooms on to a given array of mobjects (or a singular mobject)
        and automatically resizes to frame all the mobjects.

        .. NOTE::

            This method only works when 2D-objects in the XY-plane are considered, it
            will not work correctly when the camera has been rotated.

        Parameters
        ----------
        mobjects
            The mobject or array of mobjects that the camera will focus on.

        margin
            The width of the margin that is added to the frame (optional, 0 by default).

        only_mobjects_in_frame
            If set to ``True``, only allows focusing on mobjects that are already in frame.

        animate
            If set to ``False``, applies the changes instead of returning the corresponding animation

        Returns
        -------
        Union[_AnimationBuilder, ScreenRectangle]
            _AnimationBuilder that zooms the camera view to a given list of mobjects
            or ScreenRectangle with position and size updated to zoomed position.

        """
        scene_critical_x_left = None
        scene_critical_x_right = None
        scene_critical_y_up = None
        scene_critical_y_down = None

        for m in mobjects:
            if (m == self.frame) or (
                only_mobjects_in_frame and not self.is_in_frame(m)
            ):
                # detected camera frame, should not be used to calculate final position of camera
                continue

            # initialize scene critical points with first mobjects critical points
            if scene_critical_x_left is None:
                scene_critical_x_left = m.get_critical_point(LEFT)[0]
                scene_critical_x_right = m.get_critical_point(RIGHT)[0]
                scene_critical_y_up = m.get_critical_point(UP)[1]
                scene_critical_y_down = m.get_critical_point(DOWN)[1]

            else:
                if m.get_critical_point(LEFT)[0] < scene_critical_x_left:
                    scene_critical_x_left = m.get_critical_point(LEFT)[0]

                if m.get_critical_point(RIGHT)[0] > scene_critical_x_right:
                    scene_critical_x_right = m.get_critical_point(RIGHT)[0]

                if m.get_critical_point(UP)[1] > scene_critical_y_up:
                    scene_critical_y_up = m.get_critical_point(UP)[1]

                if m.get_critical_point(DOWN)[1] < scene_critical_y_down:
                    scene_critical_y_down = m.get_critical_point(DOWN)[1]

        # calculate center x and y
        x = (scene_critical_x_left + scene_critical_x_right) / 2
        y = (scene_critical_y_up + scene_critical_y_down) / 2

        # calculate proposed width and height of zoomed scene
        new_width = abs(scene_critical_x_left - scene_critical_x_right)
        new_height = abs(scene_critical_y_up - scene_critical_y_down)

        m_target = self.frame.animate if animate else self.frame
        # zoom to fit all mobjects along the side that has the largest size
        if new_width / self.frame.width > new_height / self.frame.height:
            return m_target.set_x(x).set_y(y).set(width=new_width + margin)
        else:
            return m_target.set_x(x).set_y(y).set(height=new_height + margin)

class NeuralNetworkScene(ThreeDScene):
    """Test Scene for the Neural Network"""

    def __init__(self):
        super().__init__()# camera_class=CustomCamera)

    def make_camera_follow_forward_pass(self, neural_network, buffer=1.0):
        per_layer_animations = neural_network.make_forward_pass_animation(
            per_layer_animations=True
        )
        all_layers = neural_network.all_layers
        # Compute the width and height of the frame
        """
        max_width = 0
        max_height = 0
        for layer_index in range(1, len(all_layers) - 1):
            prev_layer = all_layers[layer_index - 1]
            current_layer = all_layers[layer_index]
            next_layer = all_layers[layer_index + 1]
            group = Group(prev_layer, current_layer, next_layer)

            max_width = max(max_width, group.width)
            max_height = max(max_height, group.height)

        frame_width = max_width * (1 + buffer)
        frame_height = max_height * (1 + buffer)
        """
        animations = []
        # Start off zoomed into first section
        print(all_layers[0])
        """
        animations.append(
            FadeIn(
                Dot(
                    all_layers[0].get_center()
                )
            )
        )
        """
        animations.append(
            Group(*self.mobjects).animate.shift(
                ORIGIN - all_layers[0].get_center(),
            ),
        )
        print(all_layers[0].get_center())
        """
        zoom_animation = AnimationGroup(
            self.camera.auto_zoom(
                [
                    all_layers[0],
                ],
                margin=buffer,
            ).build(),
            run_time=5.0,
        )
        """
        #animations.append(zoom_animation)
        forward_pass_animation = per_layer_animations[all_layers[0]]
        print(per_layer_animations)
        animations.append(
            forward_pass_animation
        )
        # Go through each animation
        for layer_index in range(1, len(all_layers) - 1):
            prev_layer = all_layers[layer_index - 1]
            current_layer = all_layers[layer_index]
            next_layer = all_layers[layer_index + 1]
            # Get the current layer animation
            layer_animation = per_layer_animations[prev_layer]
            # Zoom in on a group
            """
            zoom_animation = AnimationGroup(
                self.camera.auto_zoom(
                    [
                        prev_layer, 
                        current_layer,
                        next_layer
                    ],
                    margin=buffer,
                ).build(),
                run_time=5.0,
            )
            """
            animations.append(
                Group(*self.mobjects).animate.shift(
                    ORIGIN - current_layer.get_center(),
                    run_time=5.0
                )
            )
            # animations.append(zoom_animation)
            # Now add the actual layer animation
            animations.append(layer_animation)
        return Succession(
            *animations,
            lag_ratio=1.0
        )


    def construct(self):
        # Make the Layer object
        image = Image.open("../assets/mnist/digit.jpeg")
        numpy_image = np.asarray(image)
        layer_dict = {
            "image1": ImageLayer(numpy_image, height=1.5),
            "conv1": Convolutional2DLayer(1, 7, filter_spacing=0.32),
            "conv2": Convolutional2DLayer(3, 5, 3, filter_spacing=0.32),
            "conv3": Convolutional2DLayer(5, 3, 3, filter_spacing=0.18),
            "feed1": FeedForwardLayer(3),
            "feed2": FeedForwardLayer(3),
        }
        nn = NeuralNetwork(
            layer_dict,
            layer_spacing=0.25,
            debug_mode=True
        )
        # Make Animation
        self.add(nn)
        # self.play(Create(nn))
        self.add(Dot(nn.all_layers[0].get_center(), z_index=20))
        # self.camera.frame_center = nn.all_layers[0].get_center()
        # print(self.camera.frame.get_center())
        """
        self.play(
            self.camera.frame.animate.move_to(
                nn.all_layers[0].get_center()
            )
        )
        """
        # print(self.mobjects)
        # self.play(Group(*self.mobjects).animate.shift(RIGHT))
        # print(self.camera._frame_center.get_center())
        image_mobject = ImageMobject(
            numpy_image, 
            image_mode="RGB"
        )
        self.add(image_mobject)
        self.play(
            self.camera._frame_center.animate.shift(
                np.array([0.2, 0.0, 0])
            )
        )
        """
        self.camera.frame_center = np.array([-0.5, 0, 0])
        """
        """
        self.begin_ambient_camera_rotation(rate=0.2, about="phi")
        self.wait(5)
        self.stop_ambient_camera_rotation()
        """
        """
        self.move_camera(
            frame_center=np.array([-3, 0, 0])
        )
        self.play(
            nn.make_forward_pass_animation(),
            run_time=10
        )
        self.play(
            self.make_camera_follow_forward_pass(nn),
            run_time=10
        )
        """
